# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2021 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Documents are wrappers around different forms of snapshots that may be
committed to an archive. A document primarily serves two purposes: (1) it
contains the document schema (a list of column names), and (2) it supports
reading the document rows.
"""

from __future__ import annotations
from abc import ABCMeta, abstractmethod
from dataclasses import dataclass
from datetime import datetime
from typing import Dict, Iterator, List, Optional, Tuple, Union

import pandas as pd

from histore.document.schema import DocumentSchema


"""Type aliases."""
# Scalar values and dataset rows.
Scalar = Union[int, float, str, datetime]
DataRow = List[Scalar]
RowIndex = Union[Scalar, Tuple[Scalar, ...]]


# -- Input Documents ----------------------------------------------------------

class DocumentIterator(metaclass=ABCMeta):
    """Iterator over rows in a document. Each row is represented as a list of
    cell values for the document columns. Cell values are expected to be scalar
    values. In addition to the cell values, for each row the iterator provides
    access to the row position in the document (starting at 0) and the row
    index. The row index is used as the row key when merging documents into
    archives that are not keyed by a primary key.

    For most external documents the row position and the row index will be the
    same value. Documents that represent ``pandas.DataFrame`` objects, on the
    other hand, will use the value from the data frame index for each row as
    the returned row index.
    """
    def __enter__(self) -> DocumentIterator:
        """Enter method for the context manager."""
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> bool:
        """Close the document iterator when the context manager exits."""
        self.close()
        return False

    def __iter__(self) -> DocumentIterator:
        """Return object for row iteration."""
        return self

    def __next__(self) -> Tuple[int, RowIndex, DataRow]:
        """Return next row from the iterator.

        Raises a StopIteration error when the end of the document is reached.

        Returns
        -------
        tuple of int, histore.document.base.RowIndex, histore.document.base.DataRow
        """
        return self.next()

    @abstractmethod
    def close(self):
        """Release all resources that are held by the iterator."""
        raise NotImplementedError()  # pragma: no cover

    @abstractmethod
    def next(self) -> Tuple[int, RowIndex, DataRow]:
        """Read the next row in the document.

        Returns the row position, row index and the list of cell values for each
        of the document columns. Raises a StopIteration error if an attempt is
        made to read past the end of the document.

        Returns
        -------
        tuple of int, histore.document.base.RowIndex, histore.document.base.DataRow
        """
        raise NotImplementedError()  # pragma: no cover


class Document(metaclass=ABCMeta):
    """The abstract document class maintains the document schema and provides
    access to the document rows via a document iterator.
    """
    def __init__(self, columns: DocumentSchema):
        """Initialize the document schema.

        Parameters
        ----------
        columns: list of string
            List of column names. The number of values in each document row is
            expected to be the same as the number of columns and the order of
            values in each row is expected to correspond to their respective
            column in this list.
        """
        self.columns = columns

    @abstractmethod
    def close(self):
        """Signal that processing of the document is finished.

        Any resources that were created by the document (e.g., temporary files)
        can be released.
        """
        raise NotImplementedError()  # pragma: no cover

    def iterrows(self) -> Iterator[Tuple[RowIndex, DataRow]]:
        """Simulate the iterrows() function of a pandas DataFrame.

        Returns an iterator that yields pairs of row index and row value lists
        for each row in the streamed data frame.

        Returns
        -------
        iterator
        """
        with self.open() as f:
            for _, rowidx, row in f:
                yield rowidx, row

    @abstractmethod
    def open(self) -> DocumentIterator:
        """Open the document to get a iterator for the rows in the document.

        Returns
        -------
        histore.document.base.DocumentIterator
        """
        raise NotImplementedError()  # pragma: no cover

    @abstractmethod
    def to_df(self) -> pd.DataFrame:
        """Create data frame from the document rows.

        Returns
        -------
        pd.DataFrame
        """
        raise NotImplementedError()  # pragma: no cover

    @abstractmethod
    def sorted(self, keys: List[int], buffersize: Optional[float] = None) -> Document:
        """Sort the document rows based on the values in the key columns.

        Key columns are specified by their index position. Returns a new
        document.

        Parameters
        ----------
        keys: list of int
            Index position of sort columns.
        buffersize: float, default=None
            Maximum size (in bytes) of file blocks that are kept in main-memory.

        Returns
        -------
        histore.document.base.Document
        """
        raise NotImplementedError()  # pragma: no cover


class DefaultDocument(Document):
    """Document that provides default implementation for the to_df() and
    sorted() methods of the Document class.
    """
    def __init__(self, columns: DocumentSchema):
        """Initialize the document schema.

        Parameters
        ----------
        columns: list of string
            List of column names.
        """
        super(DefaultDocument, self).__init__(columns=columns)

    def to_df(self) -> pd.DataFrame:
        """Create data frame from the document rows.

        Returns
        -------
        pd.DataFrame
        """
        return document_to_df(self)

    def sorted(self, keys: List[int], buffersize: Optional[float] = None) -> Document:
        """Sort the document rows based on the values in the key columns.

        Key columns are specified by their index position. Returns a new
        document.

        Parameters
        ----------
        keys: list of int
            Index position of sort columns.
        buffersize: float, default=None
            Maximum size (in bytes) of file blocks that are kept in main-memory.

        Returns
        -------
        histore.document.base.Document
        """
        from histore.document.sort import SortEngine
        return SortEngine(buffersize=buffersize).sorted(doc=self, keys=keys)


# -- Document descriptors -----------------------------------------------------

@dataclass
class InputDescriptor:
    """Descriptor for archive snapshot input documents."""
    # Optional user-provided description for the snapshot.
    description: Optional[str] = ''
    # Timestamp when the snapshot was first valid.
    valid_time: Optional[datetime] = None
    # Optional metadata defining the action that created the snapshot.
    action: Optional[Dict] = None


# -- Helper Functions ---------------------------------------------------------

def document_to_df(doc: Document) -> pd.DataFrame:
    """Create data frame from the document rows.

    Parameters
    ----------
    doc: histore.document.base.Document
        Input document.

    Returns
    -------
    pd.DataFrame
    """
    data, index = list(), list()
    for rowid, row in doc.iterrows():
        index.append(rowid)
        data.append(row)
    return pd.DataFrame(data=data, index=index, columns=doc.columns, dtype=object)



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2021 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Factory pattern for persistent archive managers."""

from typing import Optional

from histore.archive.manager.base import ArchiveManager
from histore.archive.manager.db.base import DBArchiveManager
from histore.archive.manager.db.database import DB
from histore.archive.manager.fs import FileSystemArchiveManager


class PersistentArchiveManager(ArchiveManager):
    """Create an instance of a persistent archive manager. There currenty are
    two implementations for persistent archive manager: (i) the file-system
    archive manager, and (ii) the archive manager that maintains archive
    descriptors in a relational databaase.
    """
    def __new__(
        cls, basedir: Optional[str] = None, dbconnect: Optional[str] = None,
        create: Optional[bool] = False
    ):
        """Create an instance of a persistent archive manager. If the database
        connector string is given an instance of the DBArchiveManager is
        returned. Otherwise, an instance of the FileSystemArchiveManager is
        returned

        Parameters
        ----------
        basedir: string, default=None
            Path to dorectory on disk where archives are maintained.
        db: histore.archive.manager.db.database.DB, default=None
            Database connection object.
        create: bool, default=False
            Create a fresh database and delete all files in the base directory
            if True.
        """
        if dbconnect is not None:
            return DBArchiveManager(
                basedir=basedir,
                db=DB(connect_url=dbconnect),
                create=create
            )
        else:
            return FileSystemArchiveManager(basedir=basedir, create=create)



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2021 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Iterator for document rows. Provides a reader that allows to iterate over
the rows in a data frame sorted by the row key.
"""

from __future__ import annotations
from abc import ABCMeta, abstractmethod
from typing import Dict, List, Tuple, Union

from histore.document.base import DocumentIterator, RowIndex
from histore.document.row import DocumentRow
from histore.key import KeyValue, NewRow, to_key


class DocumentReader(metaclass=ABCMeta):
    """Reader for rows in a document. Reads rows in order defined by the row
    key that is used for merging the document.
    """
    def __init__(self, iterator: DocumentIterator, columns: List[int]):
        """Initialize the document row iterator and the identifiers for the
        document columns.

        Parameters
        ----------
        iterator: histore.document.base.DocumentIterator
            Iterator for rows in the document.
        columns: list of int
            List of identifier for columns in the document schema.
        """
        self.iterator = iterator
        self.columns = columns

    @abstractmethod
    def annotate(self, rowidx: RowIndex, values: Dict) -> Union[KeyValue, Tuple[KeyValue, ...]]:
        """Get key value for a document row.

        The key value may either be generated from the row index or from the
        cell values.

        Parameters
        ----------
        rowidx: histore.document.base.RowIndex
            Row index value. May be a scalar value or a tuple of scalars.
        values: dict
            Mapping of column identifier to cell values.

        Returns
        -------
        histore.key.KeyValue or tuple of histore.key.KeyValue
        """
        raise NotImplementedError()  # pragma: no cover

    def close(self):
        """Close the associated document iterator."""
        self.iterator.close()

    def next(self) -> DocumentRow:
        """Read the next row in the document.

        Returns None if the end of the document has been reached.

        Returns
        -------
        histore.document.row.DocumentRow
        """
        try:
            rowpos, rowidx, values = next(self.iterator)
        except StopIteration:
            return None
        values = {colid: val for colid, val in zip(self.columns, values)}
        key = self.annotate(rowidx, values)
        return DocumentRow(pos=rowpos, key=key, values=values)


class AnnotatedDocumentReader(DocumentReader):
    """The annotated document reader uses the values from one or more document
    columns to generate keys for document rows.
    """
    def __init__(self, iterator: DocumentIterator, columns: List[int], keys: List[int]):
        """Initialize the document row iterator, the identifiers for the
        document columns, and the identifier for the key columns.

        Parameters
        ----------
        iterator: histore.document.base.DocumentIterator
            Iterator for rows in the document.
        columns: list of int
            List of identifier for columns in the document schema.
        keys: list of int
            List of identifier for document key columns.
        """
        super(AnnotatedDocumentReader, self).__init__(iterator=iterator, columns=columns)
        self.keys = keys

    def annotate(self, rowidx: RowIndex, values: Dict) -> Union[KeyValue, Tuple[KeyValue, ...]]:
        """Get key value for a document row.

        The default document reader uses the row index as the row key..

        Parameters
        ----------
        rowidx: histore.document.base.RowIndex
            Row index value. May be a scalar value or a tuple of scalars.
        values: dict
            Mapping of column identifier to cell values.

        Returns
        -------
        histore.key.KeyValue or tuple of histore.key.KeyValue
        """
        if len(self.keys) == 1:
            return to_key(values[self.keys[0]])
        return tuple([to_key(values[c]) for c in self.keys])


class DefaultDocumentReader(DocumentReader):
    """The default document reader uses the row index as the row key."""
    def __init__(self, iterator: DocumentIterator, columns: List[int]):
        """Initialize the document row iterator and the identifiers for the
        document columns.

        Parameters
        ----------
        iterator: histore.document.base.DocumentIterator
            Iterator for rows in the document.
        columns: list of int
            List of identifier for columns in the document schema.
        """
        super(DefaultDocumentReader, self).__init__(iterator=iterator, columns=columns)

    def annotate(self, rowidx: RowIndex, values: Dict) -> Union[KeyValue, Tuple[KeyValue, ...]]:
        """Get key value for a document row.

        The default document reader uses the row index as the row key..

        Parameters
        ----------
        rowidx: histore.document.base.RowIndex
            Row index value. May be a scalar value or a tuple of scalars.
        values: dict
            Mapping of column identifier to cell values.

        Returns
        -------
        histore.key.KeyValue or tuple of histore.key.KeyValue
        """
        if isinstance(rowidx, tuple) or isinstance(rowidx, list):
            return tuple([to_key(v) for v in rowidx])
        return to_key(rowidx if rowidx != -1 else NewRow())



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2021 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Compact object serializer for dataset archives. Modifies the default serializer
to save storage space by allowing for more compact serializations of timestamps
and single-version values.

Warning: This serializer CANNOT be used for datasets that have cell values that
are dictionaries.
"""

from typing import Any, List, Optional

from histore.archive.row import ArchiveRow
from histore.archive.timestamp import Timestamp
from histore.archive.value import ArchiveValue, MultiVersionValue, SingleVersionValue
from histore.archive.serialize.default import DefaultSerializer

import histore.key as anno


class CompactSerializer(DefaultSerializer):
    """Implementation of the archive object serializer. This serializer is
    intended for producing more compact serializations compared to the default
    serializer used by HISTORE.

    The compact serializer takes advantage of some optimization opportunities
    over the default serializer. It inherits from the default serializer and
    modifies the way how rows, timestamps and values are serialized.

    For rows the compact serializer takes advantage of the fact that each row
    has a fixed number of elements to it. The serializer therefore uses a list
    of objects instead of a dictionary. The elements in the list are:
    1) row identifier,
    2) row key
    3) row timestamp
    4) row position
    5) a timestamp-like serialization of the column identifier for the row cells, and
    6) a list of cell values corresponding to the list of column identifier in 5).

    For timestamps intervals that start and end at the same version are
    serialized as a single integer instead of a pair of integers.

    Archive values that only have one version are not serialized as a dictionary
    but as the value itself. This has the disadvantage that values cannot be
    dictionaries as otherwise the deserialization will fail.
    """
    def __init__(
        self, timestamp: Optional[str] = 't', pos: Optional[str] = 'p',
        name: Optional[str] = 'n', value: Optional[str] = 'v',
        colid: Optional[str] = 'c', version: Optional[str] = 'v',
        valid_time: Optional[str] = 'vt', transaction_time: Optional[str] = 'tt',
        description: Optional[str] = 'd', action: Optional[str] = 'a'
    ):
        """Initialize the labels for elements used in the serialized objects.
        By default short labels are used to reduce storage overhead.

        Passing a label that starts with the reserved character '$' will result
        in a ValueError.

        Parameters
        ----------
        timestamp: string, default='t'
            Element label for timestamp objects.
        pos: string, default='p'
            Element label for objects index position values.
        name: string, default='n'
            Element label for column names.
        value: string, default='v'
            Element label for timestamped values.
        colid: string, default='c'
            Element label for column identifier.
        version: string, default='v'
            Element label for snapshot version number.
        valid_time: string, default='vt'
            Element label for snapshot valid time.
        transaction_time: string, default='tt'
            Element label for snapshot transaction time.
        description: string, default='d'
            Element label for snapshot descriptions.
        action: string, default='a'
            Element label for snapshot actions.

        Raises
        ------
        ValueError
        """
        super(CompactSerializer, self).__init__(
            timestamp=timestamp,
            pos=pos,
            name=name,
            value=value,
            colid=colid,
            version=version,
            valid_time=valid_time,
            transaction_time=transaction_time,
            description=description,
            action=action
        )

    def deserialize_row(self, obj: List) -> ArchiveRow:
        """Get archive row instance from a serialized object. Expects a
        dictionary as created by the serialize_row method.

        Parameters
        ----------
        obj: list
            Serialized archive row object.

        Returns
        -------
        histore.archive.row.ArchiveRow
        """
        rowid = obj[0]
        key = obj[1]
        ts = self.deserialize_timestamp(obj[2])
        pos = self.deserialize_value(obj=obj[3], ts=ts)
        columns = obj[4]
        values = obj[5]
        cells = dict()
        validx = 0
        for interval in columns:
            if isinstance(interval, list):
                for i in range(interval[0], interval[1] + 1):
                    cells[i] = self.deserialize_value(obj=values[validx], ts=ts)
                    validx += 1
            else:
                cells[interval] = self.deserialize_value(obj=values[validx], ts=ts)
                validx += 1
        if isinstance(key, list) or isinstance(key, tuple):
            key = tuple([anno.to_key(k) for k in key])
        else:
            key = anno.to_key(key)
        return ArchiveRow(
            rowid=rowid,
            key=key,
            pos=pos,
            cells=cells,
            timestamp=ts
        )

    def serialize_row(self, row: ArchiveRow) -> List:
        """Get serialization for an archive row object.

        Parameters
        ----------
        row: histore.archive.row.ArchiveRow
            Archive row object that is being serialized.

        Returns
        list
        dict
        """
        ts = row.timestamp
        columns = list()
        values = list()
        for colid, val in sorted(row.cells.items(), key=lambda x: x[0]):
            if columns:
                last_interval = columns[-1]
                if last_interval[1] == colid - 1:
                    last_interval[1] = colid
                else:
                    if last_interval[0] == last_interval[1]:
                        columns[-1] = last_interval[0]
                    columns.append([colid, colid])
            else:
                columns.append([colid, colid])
            values.append(self.serialize_value(value=val, ts=ts))
        return [
            row.rowid,
            row.key,
            self.serialize_timestamp(ts),
            self.serialize_value(value=row.pos, ts=ts),
            columns,
            values
        ]

    def deserialize_value(self, obj: Any, ts: Timestamp) -> ArchiveValue:
        """Get timestamped value object from a serialization object. Expects a
        dictionary for single version values or a list of dictionaries for
        multi-version values.

        Parameters
        ----------
        obj: any
            Serialized timestamped values as generatred by the serialize_value
            method.
        ts: histore.archive.timestamp.Timestamp
            Timestamp of the parent object.

        Returns
        -------
        histore.archive.value.ArchiveValue
        """
        if isinstance(obj, dict):
            return SingleVersionValue(
                value=obj[self.value],
                timestamp=self.deserialize_timestamp(obj[self.timestamp])
            )
        elif isinstance(obj, list):
            values = []
            for v in obj:
                values.append(
                    SingleVersionValue(
                        value=v[self.value],
                        timestamp=self.deserialize_timestamp(v[self.timestamp])
                    )
                )
            return MultiVersionValue(values=values)
        else:
            return SingleVersionValue(value=obj, timestamp=ts)

    def serialize_value(self, value: ArchiveValue, ts: Timestamp) -> Any:
        """Get serialization for a timestamp value.

        Depending on whether the value is a single version value or a
        multi-version value either (i) the value itself (single-value with no
        separate timestamp), (ii) a dictionary including the single version
        value and its timestamp, or (iii) or a list of dictionaries (multi-version)
        is returned.

        Parameters
        ----------
        value: histore.archive.value.ArchiveValue
            Timestamped value in a dataset archive.
        ts: histore.archive.timestamp.Timestamp
            Timestamp of the parent object.

        Returns
        -------
        any
        """
        if value.is_single_version():
            # A single version value is either serialized as a dictionary
            # containing the value and its timestamp (if it different from the
            # parent timestamp) or as the versioned value itself.
            if value.has_timestamp and not value.timestamp.is_equal(ts):
                obj = dict()
                obj[self.timestamp] = self.serialize_timestamp(value.timestamp)
                obj[self.value] = value.value
            else:
                obj = value.value
        else:
            # A multi-version value is serialized as a list of single version
            # values each with their own timestamp.
            obj = [{
                self.timestamp: self.serialize_timestamp(v.timestamp),
                self.value: v.value
            } for v in value.values]
        return obj



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2021 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Test encoders and decoders for Json objects."""

import json

from datetime import datetime


class TestEncoder(json.JSONEncoder):
    """Json encoder that handles datetime objects."""
    def default(self, obj):
        """Convert datatime to dictionary."""
        if isinstance(obj, datetime):
            return {'$dt': obj.isoformat()}
        return obj.value  # Assumes histore.key.KeyValue


def test_decoder(obj):
    """Decode objects generated by the TestEncoder. Returns datetime object
    as string instead of datetime instances.
    """
    if '$dt' in obj:
        return obj['$dt']
    return obj



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2021 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Timestamps are sequences of snapshot (version) identifier that define when
a node was part of the archived document.
"""

from __future__ import annotations
from typing import List, Optional, Union


def TimeInterval(start: int, end: Optional[int] = None):
    """Get a timestamp interval for the given parameters.

    Mainly used for test purposes.

    Parameters
    ----------
    start: int
        First version in the interval.
    end: int, optional
        Last version (inclusive) in the interval.
    """
    return start if end is None else [start, end]


class Timestamp(object):
    """A timestamp is a sequence of integer numbers represented as a list of
    integer intervals.
    """
    def __init__(self, intervals: List[Union[int, List[int]]] = None):
        """Initialize the list of time intervals.

        Parameters
        ----------
        intervals: list of histore.archive.timestamp.TimeInterval, optional
            List of time intervals.
        """
        self.intervals = intervals

    def __str__(self) -> str:
        """Get a string representation of the timestamp.

        Returns
        -------
        string
        """
        return str(self.intervals)

    def append(self, version: int) -> Timestamp:
        """Append the given version to the timestamp. Returns a new modified
        timestamp. Raises ValueError if the given value is lower or equal to
        the last version in the sequence.

        Parameters
        ----------
        version: int
            Version identifier.

        Returns
        -------
        histore.archive.timestamp.Timestamp
        """
        if self.intervals:
            last_interval = self.interval(-1)
            current_end = last_interval[-1]
            if current_end >= version:
                raise ValueError('cannot append {} to [{}]'.format(version, self))
            elif version == current_end + 1:
                interval = [last_interval[0], version]
                return Timestamp(intervals=self.intervals[:-1] + [interval])
            else:
                return Timestamp(intervals=self.intervals + [version])
        else:
            return Timestamp(intervals=[version])

    def contains(self, version: int) -> bool:
        """Returns True if the timestamp contains the given value.

        Parameters
        ----------
        version: int
            Version identifier.

        Returns
        -------
        bool
        """
        if self.intervals:
            # Search from end of interval list since the majority of requests will
            # access more recent values (i.e., database versions).
            for index in range(len(self.intervals), 0, -1):
                interval = self.interval(index - 1)
                if interval[0] <= version <= interval[-1]:
                    return True
                elif interval[-1] < version:
                    return False
        return False

    def is_empty(self) -> bool:
        """Returns True if the timestamp is empty.

        Returns
        -------
        bool
        """
        return not self.intervals

    def is_equal(self, timestamp: Timestamp) -> bool:
        """Returns True if the two timestamps are equal, i.e., represent the
        same sequence of snapshots. Given that intervals cannot be consecutive,
        equality means (i) same number of intervals, and (ii) corresponding
        intervals have same start/end values.

        Parameters
        ----------
        timestamp: histore.archive.timestamp.Timestamp

        Returns
        -------
        bool
        """
        if len(self.intervals) == len(timestamp.intervals):
            for i in range(len(self.intervals)):
                i1 = self.interval(i)
                i2 = timestamp.interval(i)
                if not i1[0] == i2[0] and not i1[-1] == i2[-1]:
                    return False
            return True
        return False

    def interval(self, index: int) -> List[int]:
        """Get the time interval at the given index position.

        Parameters
        ----------
        index: int
            Interval index position.

        Returns
        -------
        list of int
        """
        value = self.intervals[index]
        return value if isinstance(value, list) else [value]

    def last_version(self) -> int:
        """Get the last version in the timestamp. If the timestamp is empty the
        result is None.

        Returns
        -------
        int
        """
        return self.interval(-1)[-1] if self.intervals else None

    def rollback(self, version: int) -> Timestamp:
        """Remove all versions after the given version from the timestamp.

        Returns a modifed timestamp.

        Parameters
        ----------
        version: int
            Last version in the rolled back timestamp. The result will not
            contain version numbers that are greater than the given version.

        Returns
        -------
        histore.archive.timestamp.Timestamp
        """
        # Create modified list of intervals.
        intervals = list()
        for ival in self.intervals:
            int_start, int_end = (ival[0], ival[-1]) if isinstance(ival, list) else (ival, ival)
            if int_start > version:
                # Intervals are sorted. If the current interval starts after
                # the rollback version there is nothing more to be added.
                break
            if int_end <= version:
                # Add interval since it ended before the rollback version.
                intervals.append([int_start, int_end] if int_start != int_end else int_start)
            else:
                # The rollback version falls inside the interval. Clip the
                # interval and add it to the new timestamp. All following
                # intervals can be ignored.
                intervals.append([int_start, version] if int_start != version else int_start)
                break
        return Timestamp(intervals=intervals)


class SingleVersion(Timestamp):
    """A timestamp containing a single version."""
    def __init__(self, version: int):
        """Initialize the list of time intervals.

        Parameters
        ----------
        version: int
            Version number that defines the timestamp as the only element in
            the time sequence.
        """
        return super(SingleVersion, self).__init__(intervals=[version])



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2021 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Implementation of a basic archive store that maintains all data in main
memory. Archive information is not persisted.
"""

from typing import List, Optional

from histore.archive.schema import ArchiveSchema
from histore.archive.snapshot import SnapshotListing
from histore.archive.store.base import ArchiveStore
from histore.archive.store.mem.reader import BufferedReader
from histore.archive.store.mem.writer import ArchiveBuffer


class VolatileArchiveStore(ArchiveStore):
    """Archive store that keeps all archive related information in main memory.
    The store is volatile as no information is persisted on disk.
    """
    def __init__(self, primary_key: Optional[List[int]] = None):
        """Initialize the archive archive components.

        Parameters
        ----------
        primary_key: list of int, default=None
            List of identifier for primary key columns.
        """
        self.rows = list()
        self.schema = ArchiveSchema()
        self.snapshots = SnapshotListing()
        self.row_counter = 0
        self._primary_key = primary_key

    def commit(
        self, schema: ArchiveSchema, writer: ArchiveBuffer,
        snapshots: SnapshotListing
    ):
        """Commit an updated dataset archive.

        The modified components of the archive are given as the three
        arguments of this method.

        Parameters
        ----------
        schema: histore.archive.schema.ArchiveSchema
            Schema history for the new archive version.
        writer: histore.archive.store.mem.writer.ArchiveBuffer
            Buffer containing the modified archive rows.
        snapshots: histore.archive.snapshot.SnapshotListing
            Snapshot listing for the modified archive.
        """
        self.rows = writer.rows
        self.schema = schema
        self.snapshots = snapshots
        self.row_counter = writer.row_counter

    def is_empty(self):
        """True if the archive does not contain any snapshots yet.

        Returns
        -------
        bool
        """
        return self.snapshots.is_empty()

    def get_reader(self) -> BufferedReader:
        """Get the row reader for this archive.

        Returns
        -------
        histore.archive.store.mem.BufferedReader
        """
        return BufferedReader(rows=self.rows)

    def get_schema(self) -> ArchiveSchema:
        """Get the schema history for the archived dataset.

        Returns
        -------
        histore.archive.schema.ArchiveSchema
        """
        return self.schema

    def get_snapshots(self) -> SnapshotListing:
        """Get listing of all snapshots in the archive.

        Returns
        -------
        histore.archive.snapshot.SnapshotListing
        """
        return self.snapshots

    def get_writer(self) -> ArchiveBuffer:
        """Get a a new archive buffer to maintain rows for a new archive
        version.

        Returns
        -------
        histore.archive.store.mem.ArchiveBuffer
        """
        return ArchiveBuffer(row_counter=self.row_counter)

    def primary_key(self) -> List[int]:
        """Get the list of identifier for the primary key column(s).

        Returns None if the archive is not keyed by a primary key.

        Returns
        -------
        list of int
        """
        return self._primary_key

    def rollback(self, schema: ArchiveSchema, writer: ArchiveBuffer, version: int):
        """Store the archive after it has been rolled back to a previous
        version.

        The archive schema and archive writer contain the modified schema and
        rows after the rollback.

        Parameters
        ----------
        schema: histore.archive.schema.ArchiveSchema
            Schema history for the previous archive version.
        writer: histore.archive.store.mem.writer.ArchiveBuffer
            Instance of the archive writer class returned by this store that
            was used to output the rows of the previous archive version.
        version: int
            Unique version identifier for the rollback snapshot.
        """
        snapshots = self.snapshots.rollback(version=version)
        self.rows = writer.rows
        self.schema = schema
        self.snapshots = snapshots



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2021 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Abstract class that defines the interface for the archive writer."""

from abc import ABCMeta, abstractmethod

from histore.key import NumberKey
from histore.archive.row import ArchiveRow
from histore.archive.timestamp import SingleVersion
from histore.archive.value import SingleVersionValue
from histore.document.row import DocumentRow


class ArchiveWriter(metaclass=ABCMeta):
    """The abstract archive writer class defines the methods for adding rows
    to an output archive. The writer maintains an internal counter to assign
    unique identifier to document rows.
    """
    def __init__(self, row_counter: int):
        """Initialize the counter that is used to generate unique row
        identifier.

        Parameters
        ----------
        row_counter: int
            Counter that is used to generate unique internal row identifier.
            The current value of the counter is the value for the next unique
            identifier.
        """
        self.row_counter = row_counter

    @abstractmethod
    def write_archive_row(self, row: ArchiveRow):
        """Add the given row to a new archive version.

        Parameters
        ----------
        row: histore.archive.row.ArchiveRow
            Row in a new version of a dataset archive.
        """
        raise NotImplementedError()  # pragma: no cover

    def write_document_row(self, row: DocumentRow, version: int):
        """Add a given document row to a new archive version with the given
        identifier. Uses the internal row counter to assign a new unique row
        identifier.

        Parameters
        ----------
        row: histore.document.row.DocumentRow
            Row from an input document (snapshot) that is being added to the
            archive snapshot for the given version.
        version: int
            Unique identifier for the snapshot version that the document row is
            added to.
        """
        # Create a new archive row with unique row identifier for the given
        # document row.
        ts = SingleVersion(version=version)
        cells = dict()
        for colid, value in row.values.items():
            cells[colid] = SingleVersionValue(value=value, timestamp=ts, has_timestamp=False)
        key = row.key
        if not isinstance(key, tuple) and key.is_new():
            key = NumberKey(self.row_counter)
        arch_row = ArchiveRow(
            rowid=self.row_counter,
            key=key,
            pos=SingleVersionValue(value=row.pos, timestamp=ts, has_timestamp=False),
            cells=cells,
            timestamp=ts
        )
        # Increment the row counter and add the new row to the archive.
        self.row_counter += 1
        self.write_archive_row(arch_row)


class ValidatingArchiveWriter(ArchiveWriter):
    """Wrapper for another archive writer. Ensures that archive rows are
    ordered in descending order of their key values. Raises a ValueError
    if an attempt is made to write rows that are not in order.
    """
    def __init__(self, writer: ArchiveWriter):
        """Initialize the wrapped archive writer that is used to write archive
        rows.

        Parameters
        ----------
        writer: histore.archive.writer.ArchiveWriter
            Writer for archive rows.
        """
        self.writer = writer
        self._prev = None

    @property
    def row_counter(self) -> int:
        """Counter for archive rows.

        Returns
        -------
        int
        """
        return self.writer.row_counter

    @row_counter.setter
    def row_counter(self, val: int):
        """Update the row counter for the associated writer.

        Parameters
        ----------
        val: int
            Updated row counter value.
        """
        self.writer.row_counter = val

    def write_archive_row(self, row: ArchiveRow):
        """Add the given row to a new archive version.

        Raises a ValueError if the row key is lower than the key of the
        previous row that was written.

        Parameters
        ----------
        row: histore.archive.row.ArchiveRow
            Row in a new version of a dataset archive.
        """
        if self._prev is not None and row.key < self._prev:
            msg = "row {} with key '{}' is out of order for previous key '{}'"
            raise ValueError(msg.format(row.rowid, row.key, self._prev))
        self._prev = row.key
        self.writer.write_archive_row(row)



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2021 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Abstract class for external documents that maintain data in a file on the
local file system.
"""

import os
import pandas as pd

from histore.document.base import Document, document_to_df
from histore.document.schema import DocumentSchema


class ExternalDocument(Document):
    """External document that maintains data in a file in disk."""
    def __init__(self, columns: DocumentSchema, filename: str, delete_on_close: bool):
        """Initialize the document schema and the reference to the data file.

        Parameters
        ----------
        columns: list of string
            List of column names.
        filename: string
            Path to the data file.
        delete_on_close: bool
            If True, delete the file when the document is closed.
        """
        super(ExternalDocument, self).__init__(columns=columns)
        self.filename = filename
        self.delete_on_close = delete_on_close

    def close(self):
        """Delete the file if the ``delete_on_close`` Flag is True."""
        if self.delete_on_close:
            os.unlink(self.filename)

    def to_df(self) -> pd.DataFrame:
        """Create data frame from the document rows.

        Returns
        -------
        pd.DataFrame
        """
        return document_to_df(self)



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2021 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Reader for lists of objects that are maintained as a serialized Json array
in a file on the file system.
"""

from datetime import datetime
from dateutil.parser import isoparse
from typing import Any, Callable, Dict, Optional, Tuple

import json
import os

from histore.document.base import DataRow, DocumentIterator, RowIndex

import histore.util as util


class JsonReader(object):
    """Abstract reader for rows in a Json file. Returns rows that are Json
    objects. The reader is agnostice to the format of these rows. The format is
    implementation dependent and different sub-classes will implement their own
    deserialization function.
    """
    def __init__(
        self, filename: str, compression: Optional[str] = None,
        decoder: Optional[Callable] = None
    ):
        """Initialize the input file, the row deserializer, and the Json
        decoder.

        Parameters
        ----------
        filename: string
            Path to the input file.
        compression: string, default=None
            String representing the compression mode for the output file.
        decoder: func, default=None
            Custom decoder function when reading archive rows from file.
        """
        self.decoder = decoder  # if decoder is not None else default_decoder
        if os.path.isfile(filename):
            self.fin = util.inputstream(filename, compression=compression)
        else:
            self.fin = None

    def __iter__(self):
        """Return object for row iteration."""
        return self

    def close(self):
        """Release all reseources that are associated with the reader."""
        self.fin = None

    def __next__(self) -> Any:
        """Read the next row in the file.

        Raises a StopIteration error if an attempts is made to read past the
        end of the file.

        Returns
        -------
        any
        """
        try:
            return json.loads(next(self.fin), object_hook=self.decoder)
        except TypeError:
            raise StopIteration()


class JsonIterator(DocumentIterator):
    """Document iterator for documents that have been serialized as Json
    documents. The iterator expects a file that contains a list if lists (rows).
    The first row is expected to contain the document schema. Data rows are
    expected to be triples of row position, row index, and cell values.
    """
    def __init__(
        self, filename: str, compression: Optional[str] = None,
        decoder: Optional[Callable] = None
    ):
        """Initialize the input file, the row deserializer, and the Json
        decoder.

        Parameters
        ----------
        filename: string
            Path to the data file.
        compression: string, default=None
            String representing the compression mode for the output file.
        decoder: func, default=None
            Custom decoder function when reading archive rows from file. If not
            given, the default decoder will be used.
        """
        self.reader = JsonReader(
            filename=filename,
            compression=compression,
            decoder=decoder
        )
        # Skip the column name row.
        try:
            next(self.reader)
        except StopIteration:
            pass

    def close(self):
        """Close the associated Json reader."""
        self.reader.close()

    def next(self) -> Tuple[int, RowIndex, DataRow]:
        """Read the next row in the document.

        Returns the row position, row index and the list of cell values for each
        of the document columns. Raises a StopIteration error if an attempt is
        made to read past the end of the document.

        Returns
        -------
        tuple of int, histore.document.base.RowIndex, histore.document.base.DataRow
        """
        rowpos, rowidx, values = next(self.reader)
        return rowpos, rowidx, values


# -- Helper Functions ---------------------------------------------------------

def default_decoder(obj: Dict) -> Any:
    """Default Json object decoder. Accounts for date types that have been
    encoded as dictionaries.

    Parameters
    ----------
    obj: dict
        Json object that is being encountered by the Json reader.

    Returns
    -------
    any
    """
    if '$datetime' in obj:
        return isoparse(obj['$datetime'])
    elif '$date' in obj:
        return isoparse(obj['$date']).date()
    elif '$time' in obj:
        val = obj['$time']
        if '.' in val:
            return datetime.strptime(val, '%H:%M:%S.%f').time()
        else:
            return datetime.strptime(val, '%H:%M:%S').time()
    return obj



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2021 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Columns in histore data frames have a unique identifier and a column name.
The column class extends the Python String class to be able to be used as a
column value in a Pandas data frame.
"""

from typing import List, Optional, Tuple, Union

import pandas as pd


class Column(str):
    """Columns in histore data frames are subclasses of Python strings that
    contain a unique column identifier. This implementation is based on:
    https://bytes.com/topic/python/answers/32098-my-experiences-subclassing-string

    The order of creation is that the __new__ method is called which returns
    the object then __init__ is called.
    """
    def __new__(cls, colid: int, name: str, colidx: Optional[int] = None):
        """Initialize the String object with the given column name. Ignore the
        column identifier.

        Parameters
        ----------
        colid: int
            Unique column identifier
        name: string
            Column name
        colidx: int, default=None
            Index position of the column in a dataset schema.
        """
        return str.__new__(cls, str(name))

    def __init__(self, colid: int, name: str, colidx: Optional[int] = None):
        """Initialize the unique column identifier. The column name has already
        been initialized by the __new__ method that is called prior to the
        __init__ method.

        Parameters
        ----------
        colid: int
            Unique column identifier
        name: string
            Column name
        colidx: int, default=None
            Index position of the column in a dataset schema.
        """
        self.colid = colid
        self.colidx = colidx


"""Type alias for column lists."""
# Reference to a column in a dataset schema. Columns can be referenced by
# their name or index position in the dataset schema.
ColumnRef = Union[int, str]
# Reference to one or more columns in a dataset schema.
Columns = Union[ColumnRef, List[ColumnRef]]
# The schema of a document (dataset) is a list of column names.
DocumentSchema = List[Union[str, Column]]


# -- Helper methods -----------------------------------------------------------

def as_list(columns: Columns) -> List[Union[int, str, Column]]:
    """Ensure that the given columns argument is a list.

    Parameters
    ----------
    columns: int, string, or list(int or string), optional
        Single column or list of column index positions or column names.

    Returns
    -------
    list
    """
    if isinstance(columns, pd.Index):
        return list(columns)
    elif isinstance(columns, tuple):
        return list(columns)
    elif not isinstance(columns, list):
        return [columns]
    else:
        return columns


def column_index(schema: DocumentSchema, columns: Columns):
    """Get the list of column index positions in a given schema (list of
    column names). Columns are either specified by name or by index position.
    The result is a list of column index positions.

    Raises errors if invalid columns positions or unknown column names are
    provided.

    Parameters
    ----------
    schema: list(string)
        List of column names.
    columns: int, str, or list(int or str)
        List of column index positions or column names.

    Returns
    -------
    (list, list)

    Raises
    ------
    ValueError
    """
    _, colidx = select_clause(schema=schema, columns=columns)
    return colidx


def column_ref(schema: DocumentSchema, column: ColumnRef) -> Tuple[str, int]:
    """Get the column name and index position for a referenced column in the
    given schema. Columns may be referenced by their name or index. This
    function returns both, the name and the index of the referenced column.

    Raises a ValueError if an unknown column is referenced.

    Parameters
    ----------
    schema: list of string
        List of column names in a dataset schema
    column: int, string, or Column
        Reference to a column in the dataset schema.

    Returns
    -------
    tuple(string, int)
    """
    if isinstance(column, int):
        # Raise value error if the specified index is invalid
        try:
            colname = schema[column]
            colidx = column
        except IndexError as ex:
            raise ValueError(ex)
    elif isinstance(column, Column) and column.colidx is not None:
        colidx = column.colidx
        if colidx < 0 or colidx >= len(schema):
            msg = 'invalid column index <{} {} {} />'
            raise ValueError(msg.format(column, column.colid, column.colidx))
        if schema[colidx] != column:
            msg = 'column name mismatch  <{} {} {} />'
            raise ValueError(msg.format(column, column.colid, column.colidx))
        colname = column
    elif isinstance(column, str):
        colidx = -1
        for i in range(len(schema)):
            if schema[i] == column:
                colname = schema[i]
                colidx = i
                break
        # Raise value error if the column name is unknown
        if colidx == -1:
            raise ValueError('unknown column name {}'.format(column))
    else:
        raise ValueError("invalid column reference '{}'".format(column))
    return colname, colidx


def select_clause(schema: DocumentSchema, columns: Columns) -> Tuple[List[str], List[int]]:
    """Get the list of column name objects and index positions in a data frame
    for list of columns that are specified either by name or by index position.

    The result is a tuple containing two lists: the list of column objects and
    the list of column index positions.

    Raises errors if invalid columns positions or unknown column names or types
    are provided.

    Parameters
    ----------
    schema: List of string
        List of columns in a dataset schema.
    columns: int, string or list of int or string
        Single column reference or a list of column index positions or column
        names.

    Returns
    -------
    (list, list)

    Raises
    ------
    ValueError
    """
    # Ensure that columns is a list.
    columns = as_list(columns)
    # Ensure that all elements in the selected column list are names.
    column_names = list()
    column_index = list()
    for col in columns:
        colname, colidx = column_ref(schema=schema, column=col)
        column_names.append(colname)
        column_index.append(colidx)
    return column_names, column_index


def to_schema(columns: Columns) -> List[Column]:
    """Convert a list of column names to a list of column objects.

    If all column names are already of type Column no changes will occur. If
    the given list contains a mix of types an error is raised.

    Parameters
    ----------
    columns: list of string
        List of column names.

    Returns
    -------
    list of histore.document.schema.Column
    """
    has_col = False
    has_str = False
    result = list()
    for i, col in enumerate(as_list(columns)):
        if isinstance(col, Column):
            result.append(col)
            has_col = True
        else:
            result.append(Column(colid=i, name=col, colidx=i))
            has_str = True
    if has_col and has_str:
        raise ValueError('invalid column list')
    return result



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2021 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Columns in an archive schema all have unique identifier together with
timestamped names and positions.
"""
from __future__ import annotations
from collections import defaultdict
from typing import Dict, List, Optional, Tuple

from histore.archive.provenance.base import ProvOp
from histore.archive.provenance.column import DeleteColumn, InsertColumn, UpdateColumn
from histore.archive.timestamp import SingleVersion, Timestamp
from histore.archive.value import SingleVersionValue
from histore.document.schema import Column, DocumentSchema


class ArchiveColumn(object):
    """Archive columns have a unique identifier together with a timestamped
    name and schema position.
    """
    def __init__(self, identifier: int, name: str, pos: int, timestamp: Timestamp):
        """initialize the object properties.

        Parameters
        ----------
        identifier: int
            Unique column identifier
        name: histore.archive.value.ArchiveValue
            History of names for the column.
        pos: histore.archive.value.ArchiveValue
            History of schema positions for the column.
        timestamp: histore.archive.timestamp.Timestamp
            Timestamp for the full history of the column.
        """
        self.identifier = identifier
        self.name = name
        self.pos = pos
        self.timestamp = timestamp

    def __repr__(self):
        """Unambiguous string representation of the archive column.

        Returns
        -------
        string
        """
        template = (
            '<ArchiveColumn (\n'
            '\tid=%d\n'
            '\tname=%s\n'
            '\tpos=%s\n'
            '\tts=%s\n)>'
        )
        return template % (
            self.identifier,
            str(self.name),
            str(self.pos),
            str(self.timestamp)
        )

    def at_version(self, version: int) -> Tuple[str, int]:
        """Get the name and schema position for the column in the given
        version of a dataset. Returns a tuple of (name, pos).

        Parameters
        ----------
        version: int
            Unique version identifier.

        Returns
        -------
        (string, int)
        """
        return self.name.at_version(version), self.pos.at_version(version)

    def diff(self, original_version: int, new_version: int) -> ProvOp:
        """Get provenance information representing the difference for this
        column between the original version and a new version.

        The three main types of changes that the column can experience are:

        - InsertColumn: The column did not exists in the original version but
          in the new version.
        - DeleteColumn: The column did exist in the original version but not in
          the new version.
        - UpdateColumn: Column updates can be comprised of two types of
          changes: (i) The column position changed, or (ii) the column name
          changed.

        If the column has no changes between both versions the result is None.

        Parameters
        ----------
        original_version: int
            Unique identifier for the original version.
        new_version: int
            Unique identifier for the version that the original version is
            compared against.

        Returns
        -------
        histore.archive.provenance.base.ProvOp
        """
        exists_in_orig = self.timestamp.contains(original_version)
        exists_in_new = self.timestamp.contains(new_version)
        if exists_in_orig and not exists_in_new:
            # Column has been deleted
            return DeleteColumn(self.identifier)
        elif not exists_in_orig and exists_in_new:
            # Column has been inserted
            return InsertColumn(self.identifier)
        elif exists_in_orig and exists_in_new:
            # Check if column name or position has been updated.
            upd_name = self.name.diff(original_version, new_version)
            upd_pos = self.pos.diff(original_version, new_version)
            if upd_pos is not None or upd_name is not None:
                return UpdateColumn(
                    key=self.identifier,
                    name=upd_name,
                    position=upd_pos
                )
        # Column has not changed
        return None

    def merge(self, name: str, pos: int, version: int) -> ArchiveColumn:
        """Create a modified version of the column for the given version. Adds
        the name and index position to the respective history of the column.
        Returns a modified copy of the column.

        Parameters
        ----------
        name: string
            Column name.
        pos: int
            Position of the column in the snapshot schema.

        Returns
        -------
        histore.archive.schema.ArchiveColumn
        """
        return ArchiveColumn(
            identifier=self.identifier,
            name=self.name.merge(value=name, version=version),
            pos=self.pos.merge(value=pos, version=version),
            timestamp=self.timestamp.append(version)
        )

    def rollback(self, version: int) -> ArchiveColumn:
        """Rollback the versions of the column.

        Truncates the timestamps of the column to the given rollback version.
        Returns a new archive column object or None if the column did not exist
        at or prior to the rollback version.

        Parameters
        ----------
        version: int
            Unique identifier of the rollback version.

        Returns
        -------
        histore.archive.schema.ArchiveColumn
        """
        ts = self.timestamp.rollback(version=version)
        if ts.is_empty():
            return None
        return ArchiveColumn(
            identifier=self.identifier,
            name=self.name.rollback(version=version),
            pos=self.pos.rollback(version=version),
            timestamp=ts
        )


class ArchiveSchema(object):
    """The archive schema maintains an index of columns in the history of a
    dataset. Each column has a unique identifier together with timestamped
    names and positions.
    """
    def __init__(self, columns: List[ArchiveColumn] = None):
        """Initialize the index of columns in the schema. Raises a ValueError
        if the identifier of columns are not unique.

        Parameters
        ----------
        columns: dict or list, default=None
            Dictionary or list of archive columns. In a dictionary, columns are
            indexed by their unique identifier. If a list is given it will be
            converted to a dictionary.
        """
        if columns is None:
            # Assert that columns is not None
            self.columns = dict()
        elif isinstance(columns, list):
            # If columns is a list convert it to a dictionary
            self.columns = dict()
            for c in columns:
                if c.identifier in self.columns:
                    raise ValueError('duplicate identifier %d' % c.identifier)
                self.columns[c.identifier] = c
        elif isinstance(columns, dict):
            self.columns = columns
        else:
            raise ValueError('expected list or dict of columns')

    def __iter__(self):
        """Make column list iterable."""
        return iter(self.columns.values())

    def __repr__(self):
        """Unambiguous string representation of the archive schema.

        Returns
        -------
        string
        """
        return '\n'.join(str(c) for c in self.columns.values())

    def at_version(self, version: int) -> List[Column]:
        """Get the schema for the snapshot at the given version of a dataset.
        Returns a list of snapshot columns in their original order.

        Parameters
        ----------
        version: int
            Unique version identifier.

        Returns
        -------
        list(histore.document.schema.Column)
        """
        # Get a list of (identifier, name, pos) for all columns in the snapshot
        # schema at the given version.
        cols = list()
        for col in self.columns.values():
            if col.timestamp.contains(version):
                name, pos = col.at_version(version)
                cols.append((col.identifier, name, pos))
        # Sort columns based on their position and return a list of snapshot
        # columns.
        cols.sort(key=lambda x: x[2])
        return [Column(colid=id, name=name, colidx=pos) for id, name, pos in cols]

    def merge(
        self, columns: DocumentSchema, version: int, origin: Optional[int] = None,
        renamed: Optional[List[Tuple[str, str]]] = None
    ) -> Tuple[ArchiveSchema, List[int]]:
        """Add the given snapshot columns to the schema history for a dataset.

        Expects a list of strings or identifiable column objects. If column
        objects are given their identifier must match an existing identifier
        in the schema and all identifier have to be unique. Column names that
        are given as string-only are considered to be new columns.

        When matching by name the columns in the given list are matched against
        columns in the schema of the origin version. Columns that are unmatched
        are considered new columns. If the list of columns of the origin
        schema contains duplicate column names matching by name is not possible
        and an error is raised.

        Returns the new archive schema and a list of column identifier for the
        document columns. The order in the identifier list corresponds to the
        order of columns in the document schema.

        Parameters
        ----------
        columns: list
            List of column names in their original order. Column names can
            either be strings of snapshot columns.
        version: int
            Identifier of the new version.
        origin: int
            Identifier of the previos snapshot. This identifier is required
            when matching columns by name.
        renamed: list of tuple of str, str, default=None
            Optional mapping of columns that have been renamed. Tuples contain
            the old and the new columns name.

        Returns
        -------
        histore.archive.schema.ArchiveSchema, list

        Raises
        ------
        ValueError
        """
        # Initialize the result list of identifier for document columns.
        document_columns = [None] * len(columns)
        # Create type-dependent mappings for identifiable and non-identifiable
        # columns. Identifiable columns are mapped to a tuple of position and
        # columns. Non-identifiable columns are mapped from their name to their
        # position in the document schema.
        id_cols, name_cols = document_schema(columns)
        # Create an index for the new schema columns.
        schema = dict()
        # Add identifiable columns to the new archive schema. Merge them with
        # an existing column that has the same identifier.
        for colid in id_cols:
            pos, name = id_cols[colid]
            if colid in self.columns:
                col = self.columns[colid].merge(name=name, pos=pos, version=version)
            else:
                col = create_column(colid=colid, name=name, pos=pos, version=version)
            schema[colid] = col
            document_columns[pos] = colid
        # Matched non-identifiable columns by name.
        if name_cols:
            # Create mapping of new names to old names for renamed columns.
            renamed_to = {new: old for old, new in renamed} if renamed else dict()
            # Create a mapping of column names in the schema of the previous
            # snapshot to their column identifier. Note that column names are
            # not necessarily unique. Thus, we map names to list of identifier.
            # If a document column is mapped to a previous column with multiple
            # identifier a ValueError is raised.
            prev_schema = schema_mapping(self.at_version(origin)) if origin is not None else dict()
            # Counter for new column identifier.
            colids = list(id_cols.keys()) + list(self.columns.keys())
            col_counter = max(colids) + 1 if colids else 0
            # Match named columns. Columns that are unmatched are considered
            # new columns.
            for name, pos in name_cols.items():
                colname = renamed_to.get(name, name)
                matches = prev_schema.get(colname, [])
                if len(matches) > 1:
                    raise ValueError("cannot match '{}' to unique column".format(name))
                if len(matches) == 1:
                    # Found matching column.
                    colid = matches[0]
                    col = self.columns[colid].merge(name=name, pos=pos, version=version)
                else:
                    # Create a new column.
                    colid = col_counter
                    col = create_column(colid=colid, name=name, pos=pos, version=version)
                    col_counter += 1
                if colid in schema:
                    # Raise an error if a archive column is matched to multiple
                    # document columns.
                    raise ValueError("multiple matches for {}".format(colid))
                schema[colid] = col
                document_columns[pos] = colid
        # Add remaining un-matched archive columns to the new schema.
        for colid, col in self.columns.items():
            if colid not in schema:
                schema[colid] = col
        return ArchiveSchema(columns=list(schema.values())), document_columns

    def rollback(self, version: int) -> ArchiveSchema:
        """Rollback timestamps of the archive schema.

        Truncates the timestamp of all columns. Returns a new schema containing
        only those columns that existed prior or at the given rollback version.
        the result may be an empty schema but never None.

        Parameters
        ----------
        version: int
            Unique identifier of the rollback version.

        Returns
        -------
        histore.archive.schema.ArchiveSchema
        """
        columns = dict()
        for colid, col in self.columns.items():
            col = col.rollback(version=version)
            if col is not None:
                columns[colid] = col
        return ArchiveSchema(columns=columns)


# -- Helper Methods -----------------------------------------------------------

def create_column(colid: int, name: str, pos: int, version: int) -> ArchiveColumn:
    """Create a new archive column object for a document schema column.

    Parameters
    ----------
    colid: int
        Unique column identifier.
    name: str
        Column name.
    pos: int
        Column positon in the document schema.
    version: int
        Snapshot version identifier.

    Returns
    -------
    histore.archive.schema.ArchiveColumn
    """
    ts = SingleVersion(version=version)
    return ArchiveColumn(
        identifier=colid,
        name=SingleVersionValue(value=name, timestamp=ts),
        pos=SingleVersionValue(value=pos, timestamp=ts),
        timestamp=ts
    )


def document_schema(columns: DocumentSchema) -> Tuple[Dict[int, Tuple[int, str]], Dict[str, int]]:
    """Create type-dependent mappings for document schema columns.

    For columns that are identifiable a mapping from the column identifier to
    tuples of column objects and their index position is created. A ValueError
    is raised if the identifier is not unique.

    For columns that are not identifiable a second mapping is created that maps
    their name to their index position. Raises a ValueError if the names of
    non-identifiable columns are not unique.

    Parameters
    ----------
    columns: list
        List of column names in their original order. Column names can
        either be strings of snapshot columns.

    Returns
    -------
    tuple of dict, dict
    """
    id_cols, name_cols = dict(), dict()
    for colpos, col in enumerate(columns):
        if isinstance(col, Column):
            colid = col.colid
            if colid in id_cols:
                raise ValueError("duplicate column identifier '{}'".format(colid))
            id_cols[colid] = (colpos, col)
        else:
            if col in name_cols:
                raise ValueError("duplicate column '{}'".format(col))
            name_cols[col] = colpos
    return id_cols, name_cols


def schema_mapping(columns: List[Column]) -> Dict(str, List):
    """Create a mapping of column names in the schema of the previous snapshot
    to their column identifier.

    Column names are mapped to lists of identifiers since column names do not
    have to be unique.

    Parameters
    ----------
    columns: list of histore.document.schema.Column
        Columns in a document schema.

    Returns
    -------
    dict
    """
    prev_schema = defaultdict(list)
    for col in columns:
        prev_schema[col].append(col.colid)
    return prev_schema



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2021 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Reader for archive rows that are maintained as a Json array in a file on the
file system.
"""

from typing import Callable, Optional

from histore.archive.reader import ArchiveReader
from histore.archive.serialize.base import ArchiveSerializer
from histore.archive.serialize.default import DefaultSerializer
from histore.document.json.reader import JsonReader


class ArchiveFileReader(ArchiveReader):
    """Reader for rows in a dataset archive. Reads rows in ascending order of
    their identifier.
    """
    def __init__(
        self, filename: str, serializer: Optional[ArchiveSerializer] = None,
        compression: Optional[str] = None, decoder: Optional[Callable] = None
    ):
        """
        Parameters
        ----------
        filename: string
            Path to the output file.
        serializer: histore.archive.serialize.base.ArchiveSerializer, default=None
            Implementation of the archive serializer interface that is used to
            serialize rows that are written to file.
        compression: string, default=None
            String representing the compression mode for the output file.
        decoder: func, default=None
            Custom decoder function when reading archive rows from file. If not
            given, the default decoder will be used.
        """
        self.serializer = serializer if serializer else DefaultSerializer()
        # Create Json reader for the input file.
        self.reader = JsonReader(
            filename=filename,
            compression=compression,
            decoder=decoder
        )

    def close(self):
        """Release all reseources that are associated with the reader."""
        self.reader.close()

    def next(self):
        """Read the next row in the dataset archive. Returns None if the end of
        the archive rows has been reached.

        Returns
        -------
        histore.archive.row.ArchiveRow
        """
        try:
            return self.serializer.deserialize_row(next(self.reader))
        except StopIteration:
            self.close()
            return None



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2021 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Factory pattern for archives that are maintained on disk."""

from typing import Callable, Dict, Optional, Union

import json
import os
import shutil

from histore.archive.base import Archive, InputDocument, PersistentArchive, PrimaryKey
from histore.archive.manager.base import ArchiveManager
from histore.archive.manager.descriptor import ArchiveDescriptor
from histore.archive.store.fs.base import ArchiveFileStore
from histore.document.base import InputDescriptor

import histore.config as config
import histore.util as util


class FileSystemArchiveManager(ArchiveManager):
    """The persistent archive manager maintains a set of archives on disk. The
    list of archive descriptors is also maintained on disk as a Json file.
    """
    def __init__(
        self, basedir: Optional[str] = None, create: Optional[bool] = False
    ):
        """Initialize the base directory under which all archives are stored in
        individual sub-folders. If the base directory is not given the value
        will be read from the environment variable HISTORE_BASEDIR or the
        default value $HOME/.histore.

        Read the Json file containing the archive descriptors (if it exist).

        Parameters
        ----------
        basedir: string
            Path to dorectory on disk where archives are maintained.
        create: bool, default=False
            Create a fresh instance of the archive manager if True. This will
            delete all files in the base directory.
        """
        if basedir is None:
            basedir = config.BASEDIR()
        self.basedir = basedir
        # Initialize path to file that maintains archive descriptors.
        if create:
            # Clear all files in the base directory if it exists.
            if os.path.isdir(self.basedir):
                util.cleardir(self.basedir)
        util.createdir(self.basedir)
        # Initialize the internal cache of archive descriptors
        self.descriptorfile = os.path.join(basedir, 'archives.json')
        self._archives = dict()
        if os.path.isfile(self.descriptorfile):
            with open(self.descriptorfile, 'r') as f:
                doc = json.load(f)
            for obj in doc:
                descriptor = ArchiveDescriptor(obj)
                self._archives[descriptor.identifier()] = descriptor

    def _archive_dir(self, identifier: str) -> str:
        """Get directory for archive with the given identifier.

        Parameters
        ----------
        identifier: string
            Unique archive identifier.

        Returns
        -------
        string
        """
        return os.path.join(self.basedir, identifier)

    def archives(self) -> Dict[str, ArchiveDescriptor]:
        """Get dictionary of archive descriptors. The returned dictionary maps
        archive identifier to their descriptor.

        Returns
        -------
        dict(string: histore.archive.manager.descriptor.ArchiveDescriptor)
        """
        return self._archives

    def create(
        self, name: Optional[str] = None, description: Optional[str] = None,
        encoder: Optional[str] = None, decoder: Optional[str] = None,
        serializer: Union[Dict, Callable] = None, doc: Optional[InputDocument] = None,
        primary_key: Optional[PrimaryKey] = None, snapshot: Optional[InputDescriptor] = None,
        sorted: Optional[bool] = False, buffersize: Optional[float] = None,
        validate: Optional[bool] = False
    ) -> ArchiveDescriptor:
        """Create a new archive object under a given unique name.

        For archives that are keyed by a primary key, the input document for
        the first dataset snapshot has to be provided. This snapshot will be
        loaded into the archive.

        Raises a ValueError if an archive with the given name exists.

        Parameters
        ----------
        name: string, default=None
            Descriptive name that is associated with the archive.
        description: string, default=None
            Optional long description that is associated with the archive.
        encoder: string, default=None
            Full package path for the Json encoder class that is used by the
            persistent archive.
        decoder: string, default=None
            Full package path for the Json decoder function that is used by the
            persistent archive.
        serializer: dict or callable, default=None
            Dictionary or callable that returns a dictionary that contains the
            specification for the serializer. The serializer specification is
            a dictionary with the following elements:
            - ``clspath``: Full package target path for the serializer class
            that is instantiated.
            - ``kwargs`` : Additional arguments that are passed to the
            constructor of the created serializer instance.
            Only ``clspath`` is required.
        doc: histore.archive.base.InputDocument, default=None
            Input document representing the initial dataset snapshot that is
            being loaded into the archive.
        primary_key: string or list, default=None
            Column(s) that are used to generate identifier for snapshot rows.
        snapshot: histore.document.base.InputDescriptor, default=None
            Optional metadata for the created snapshot.
        sorted: bool, default=False
            Flag indicating if the document is sorted by the optional primary
            key attributes. Ignored if the archive is not keyed.
        buffersize: float, default=None
            Maximum size (in bytes) for the memory buffer when sorting the
            input document.
        validate: bool, default=False
            Validate that the resulting archive is in proper order before
            committing the action.

        Returns
        -------
        histore.archive.manager.descriptor.ArchiveDescriptor

        Raises
        ------
        ValueError
        """
        # Ensure that the archive name is unique.
        if self.get_by_name(name) is not None:
            raise ValueError("archive '{}' already exists".format(name))
        # Create an unique identifier for the new archive.
        identifier = util.get_unique_identifier()
        # Create persistent archive. This will load the initial snapshot if a
        # document is given.
        archive = PersistentArchive(
            basedir=self._archive_dir(identifier),
            create=True,
            serializer=serializer,
            encoder=encoder,
            decoder=decoder,
            descriptor=snapshot,
            doc=doc,
            primary_key=primary_key,
            sorted=sorted,
            buffersize=buffersize,
            validate=validate
        )
        # Create the descriptor for the new archive.
        descriptor = ArchiveDescriptor.create(
            identifier=identifier,
            name=name,
            description=description,
            primary_key=archive.store.primary_key(),
            encoder=encoder,
            decoder=decoder,
            serializer=serializer
        )
        # Write list of archive descriptors.
        self._archives[identifier] = descriptor
        self.write()
        return descriptor

    def delete(self, identifier: str):
        """Delete the archive with the given identifier.

        Parameters
        ----------
        identifier: string
            Unique archive identifier
        """
        if self.contains(identifier):
            archdir = self._archive_dir(identifier)
            shutil.rmtree(archdir)
            del self._archives[identifier]
            self.write()

    def get(self, identifier: str) -> Archive:
        """Get the archive that is associated with the given identifier. Raises
        a ValueError if the identifier is unknown.

        Parameters
        ----------
        identifier: string
            Unique archive identifier

        Returns
        -------
        histore.archive.base.Archive

        Raises
        ------
        ValueError
        """
        desc = self._archives.get(identifier)
        if desc is None:
            raise ValueError('unknown archive {}'.format(identifier))
        store = ArchiveFileStore(
            basedir=self._archive_dir(identifier),
            replace=False,
            serializer=desc.serializer(),
            encoder=desc.encoder(),
            decoder=desc.decoder(),
            primary_key=desc.primary_key()
        )
        return Archive(store=store)

    def rename(self, identifier: str, name: str):
        """Rename the specified archive. Raises a ValueError if the identifier
        is unknown or if an archive with the given name exist.

        Parameters
        ----------
        identifier: string
            Unique archive identifier
        name: string
            New archive name.

        Raises
        ------
        ValueError
        """
        archive = self._archives.get(identifier)
        if archive is None:
            raise ValueError("unknown archive '{}''".format(identifier))
        if archive.name() == name:
            # Do nothing if the archive name matches the new name.
            return
        # Raise an error if another archive with the new name exists.
        # Ensure that the archive name is unique.
        if self.get_by_name(name) is not None:
            raise ValueError("archive '{}' already exists".format(name))
        archive.rename(name)
        # Write list of archive descriptors.
        self.write()

    def write(self):
        """Write the current descriptor set to file."""
        with open(self.descriptorfile, 'w') as f:
            json.dump([d.doc for d in self._archives.values()], f)



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2021 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Store for archives that are materialized as files on the file system. Each
archive is maintained in a separate folder on disk. For each archive two
different files are created.

- metadata.dat: Json file containing the archive schema, snapshot information,
  and the row counter.
- rows.dat: Data fie containing the archive rows.
"""

from typing import Callable, List, Optional

import json
import os
import shutil

from histore.archive.schema import ArchiveSchema
from histore.archive.serialize.default import ArchiveSerializer, DefaultSerializer
from histore.archive.snapshot import SnapshotListing
from histore.archive.store.base import ArchiveStore
from histore.archive.store.fs.reader import ArchiveFileReader
from histore.archive.store.fs.writer import ArchiveFileWriter
from histore.document.json.reader import default_decoder
from histore.document.json.writer import DefaultEncoder

import histore.util as util


"""Element labels for metadata serialization."""
META_PRIMARYKEY = 'primaryKey'
META_SCHEMA = 'schema'
META_SNAPSHOTS = 'snapshots'
META_ROWCOUNT = 'rowcount'


class ArchiveFileStore(ArchiveStore):
    """Archive store that maintains archive data in files on the file system.
    Maintains a copy of the schema and the snapshot descriptors in memory for
    faster access.
    """
    def __init__(
        self, basedir: str, primary_key: Optional[List[int]] = None,
        replace: Optional[bool] = False, serializer: Optional[ArchiveSerializer] = None,
        compression: Optional[str] = None, encoder: Optional[json.JSONEncoder] = None,
        decoder: Optional[Callable] = None
    ):
        """Initialize the archive archive components.

        Parameters
        ----------
        basedir: string
            Path to the base directory for archive files. If the directory does
            not exist it will be created.
        primary_key: list of int, default=None
            List of identifier for primary key columns.
        replace: boolean, default=False
            Do not read existing files in the base directory to initialize the
            archive. Treat the archive as being empty instead if True.
        serializer: histore.archive.serialize.base.ArchiveSerializer, default=None
            Implementation of the archive serializer interface that is used to
            serialize rows that are written to file.
        compression: string, default=None
            String representing the compression mode. Only te data file will be
            compressed. the metadata file is always storesd as plain text.
        encoder: json.JSONEncoder, default=None
            Encoder used when writing archive rows as JSON objects to file.
        decoder: func, default=None
            Custom decoder function when reading archive rows from file.
        """
        self.basedir = util.createdir(basedir)
        self._primary_key = primary_key
        self.serializer = serializer if serializer else DefaultSerializer()
        self.compression = compression
        self.encoder = encoder if encoder is not None else DefaultEncoder
        self.decoder = decoder if decoder is not None else default_decoder
        # Initialize the file names
        self.datafile = os.path.join(self.basedir, 'rows.dat')
        self.metafile = os.path.join(self.basedir, 'metadata.dat')
        self.tmpdatafile = os.path.join(self.basedir, 'rows.tmp')
        self.tmpmetafile = os.path.join(self.basedir, 'metadata.tmp')
        if not replace and os.path.isfile(self.metafile):
            # Read schema and snapshot information from disk if the metadata
            # file exists.
            with open(self.metafile, 'r') as f:
                doc = json.load(f, object_hook=self.decoder)
            # Deserialize schema columns.
            columns = list()
            for c in doc.get(META_SCHEMA, []):
                columns.append(self.serializer.deserialize_column(c))
            self.schema = ArchiveSchema(columns=columns)
            # Deserialize snapshot descriptors
            snapshots = list()
            for s in doc.get(META_SNAPSHOTS, []):
                snapshots.append(self.serializer.deserialize_snapshot(s))
            self.snapshots = SnapshotListing(snapshots=snapshots)
            # Deserialize row counter.
            self.row_counter = doc.get(META_ROWCOUNT, 0)
            # Overwrite primary key if it is present in the metadata file.
            if self._primary_key:
                doc[META_PRIMARYKEY] = self._primary_key
                with open(self.tmpmetafile, 'w') as f:
                    json.dump(doc, f, cls=self.encoder)
            else:
                self._primary_key = doc[META_PRIMARYKEY]
        else:
            # Create an empty archive.
            self.schema = ArchiveSchema()
            self.snapshots = SnapshotListing()
            self.row_counter = 0
            # Remove any previous files that may exist in the base folder.
            for f in [self.datafile, self.metafile]:
                if os.path.isfile(f):
                    os.remove(f)
            # Write primary key to metadata file.
            with open(self.metafile, 'w') as f:
                json.dump({META_PRIMARYKEY: self._primary_key}, f)

    def commit(
        self, schema: ArchiveSchema, writer: ArchiveFileWriter,
        snapshots: SnapshotListing
    ):
        """Commit an updated dataset archive.

        The modified components of the archive are given as the three
        arguments of this method.

        Parameters
        ----------
        schema: histore.archive.schema.ArchiveSchema
            Schema history for the new archive version.
        writer: histore.archive.store.fs.writer.ArchiveFileWriter
            Instance of the archive writer class returned by this store that
            was used to output the rows of the new archive version.
        snapshots: histore.archive.snapshot.SnapshotListing
            Snapshot listing for the modified archive.
        """
        # Materialize the modified archive.
        self._write(schema=schema, writer=writer, snapshots=snapshots)
        # Update the cached objects
        self.schema = schema
        self.snapshots = snapshots
        self.row_counter = writer.row_counter

    def is_empty(self) -> bool:
        """True if the archive does not contain any snapshots yet.

        Returns
        -------
        bool
        """
        return self.snapshots.is_empty()

    def get_reader(self) -> ArchiveFileReader:
        """Get the row reader for this archive.

        Returns
        -------
        histore.archive.store.mem.BufferedReader
        """
        return ArchiveFileReader(
            filename=self.datafile,
            serializer=self.serializer,
            compression=self.compression,
            decoder=self.decoder
        )

    def get_schema(self) -> ArchiveSchema:
        """Get the schema history for the archived dataset.

        Returns
        -------
        histore.archive.schema.ArchiveSchema
        """
        return self.schema

    def get_snapshots(self) -> SnapshotListing:
        """Get listing of all snapshots in the archive.

        Returns
        -------
        histore.archive.snapshot.SnapshotListing
        """
        return self.snapshots

    def get_writer(self, validate: Optional[bool] = None) -> ArchiveFileWriter:
        """Get a a new archive buffer to maintain rows for a new archive
        version.

        Returns
        -------
        histore.archive.store.mem.ArchiveBuffer
        """
        # Use a temporary file in the base directory to write the new archive
        # before commit.
        return ArchiveFileWriter(
            filename=self.tmpdatafile,
            row_counter=self.row_counter,
            serializer=self.serializer,
            compression=self.compression,
            encoder=self.encoder
        )

    def primary_key(self) -> List[int]:
        """Get the list of identifier for the primary key column(s).

        Returns None if the archive is not keyed by a primary key.

        Returns
        -------
        list of int
        """
        return self._primary_key

    def rollback(self, schema: ArchiveSchema, writer: ArchiveFileWriter, version: int):
        """Store the archive after it has been rolled back to a previous
        version.

        The archive schema and archive writer contain the modified schema and
        rows after the rollback.

        Parameters
        ----------
        schema: histore.archive.schema.ArchiveSchema
            Schema history for the previous archive version.
        writer: histore.archive.store.fs.writer.ArchiveFileWriter
            Instance of the archive writer class returned by this store that
            was used to output the rows of the previous archive version.
        version: int
            Unique version identifier for the rollback snapshot.
        """
        # Get an updated shapshot listing.
        snapshots = self.snapshots.rollback(version=version)
        # Materialize the modified archive.
        self._write(schema=schema, writer=writer, snapshots=snapshots)
        # Update the cached objects
        self.schema = schema
        self.snapshots = snapshots

    def _write(self, schema, writer, snapshots):
        """Materialize the modified archive.

        Parameters
        ----------
        schema: histore.archive.schema.ArchiveSchema
            Schema history for the previous archive version.
        writer: histore.archive.store.fs.writer.ArchiveFileWriter
            Instance of the archive writer class returned by this store that
            was used to output the rows of the previous archive version.
        snapshots: histore.archive.snapshot.SnapshotListing
            Modified snapshot listing.
        """
        encoder = self.serializer
        doc = {
            META_SCHEMA: [encoder.serialize_column(c) for c in schema],
            META_SNAPSHOTS: [encoder.serialize_snapshot(s) for s in snapshots],
            META_ROWCOUNT: writer.row_counter,
            META_PRIMARYKEY: self._primary_key
        }
        with open(self.tmpmetafile, 'w') as f:
            json.dump(doc, f, cls=self.encoder)
        # Close the archive writer.
        writer.close()
        # Replace existing files with temporary files for new archive version.
        # This is the point of no return.
        # TODO: Instead of moving files we could delete (or keep) previous
        # files as backup.
        shutil.move(src=self.tmpmetafile, dst=self.metafile)
        shutil.move(src=self.tmpdatafile, dst=self.datafile)



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2021 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Documents that are serialized as Json arrays. The file format is a list of
lists (rows). The first row is expected to be the list of column names. All
following rows are triples of row position, row identifier and a list of cell
values for each of the columns.
"""

from typing import Callable, List, Optional

import json

from histore.document.base import Document
from histore.document.external import ExternalDocument
from histore.document.json.reader import JsonReader, JsonIterator
from histore.document.sort import SortEngine


class JsonDocument(ExternalDocument):
    """File containing a document that is serialized as a list of Json rows.
    The first row is assumed to contain the list of column names. The following
    rows are triples of row position, row identifier and a list of cell values.
    """
    def __init__(
        self, filename: str, compression: Optional[str] = None,
        encoder: Optional[json.JSONEncoder] = None, decoder: Optional[Callable] = None,
        delete_on_close: Optional[bool] = False
    ):
        """Initialize inofrmation about the file that contains the Json
        serialization of the document.

        Parameters
        ----------
        filename: string
            Path to the data file.
        compression: string, default=None
            String representing the compression mode for the output file.
        encoder: json.JSONEncoder, default=None
            Encoder used when writing rows as JSON objects to file during
            external merge sort.
        decoder: func, default=None
            Custom decoder function when reading archive rows from file. If not
            given, the default decoder will be used.
        delete_on_close: bool, default=False
            If True, delete the file when the document is closed.
        """
        self.compression = compression
        self.encoder = encoder
        self.decoder = decoder
        # Open the reader for the input file to get the list of column names
        # from the first row. If the file is empty the list of rows is empty.
        reader = JsonReader(
            filename=filename,
            compression=compression,
            decoder=decoder
        )
        try:
            columns = next(reader)
        except StopIteration:
            columns = []
        reader.close()
        # Initialize the super class.
        super(JsonDocument, self).__init__(
            columns=columns,
            filename=filename,
            delete_on_close=delete_on_close
        )

    def open(self) -> JsonIterator:
        """Open the document to get a iterator for the rows in the document.

        Returns
        -------
        histore.document.json.base.JsonIterator
        """
        return JsonIterator(
            filename=self.filename,
            compression=self.compression,
            decoder=self.decoder
        )

    def sorted(self, keys: List[int], buffersize: Optional[float] = None) -> Document:
        """Sort the document rows based on the values in the key columns.

        Key columns are specified by their index position. Returns a new
        document.

        Parameters
        ----------
        keys: list of int
            Index position of sort columns.
        buffersize: float, default=None
            Maximum size (in bytes) of file blocks that are kept in main-memory.

        Returns
        -------
        histore.document.base.Document
        """
        sort = SortEngine(buffersize=buffersize, encoder=self.encoder, decoder=self.decoder)
        return sort.sorted(doc=self, keys=keys)



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2021 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Class for sorting input documents in main-memory or using secondary storage.
The functions in this module are inspired by csvsort
(https://github.com/richardpenman/csvsort).

For HISTORE we made a few modifications that allow us to handle arbitraty input
documents. First, if the document rows can be kept in main-memory they will be
sorted directly instead of writing them to disk after sorting. Second, we use
JSON as the serialization format to retain raw data types of row values.
"""

from tempfile import NamedTemporaryFile as TempFile
from typing import Callable, List, Optional, Tuple

import heapq
import json
import os
import sys

from histore.document.base import Document, DocumentIterator
from histore.document.json.reader import JsonIterator
from histore.document.json.writer import JsonWriter
from histore.document.mem import InMemoryDocument
from histore.document.schema import DocumentSchema

import histore.config as config
import histore.util as util


class SortEngine:
    """External merge sort for input documents. Seralizes a given document into
    JSON format. The advantage of using JSON as the seriaization format is the
    ablity to maintain the data type of the document rows.
    """
    def __init__(
        self, buffersize: Optional[float] = None,
        encoder: Optional[json.JSONEncoder] = None,
        decoder: Optional[Callable] = None
    ):
        """Initialize the JSON encoder and decoder.

        Uses the default encoder and decoder if the respective argument values
        are None.

        Parameters
        ----------
        buffersize: float, default=None
            Maximum size (in bytes) of file blocks that are kept in main-memory.
            If no buffer size is given the value from the environment variable
            HISTORE_SORTBUFFER is used.
        encoder: json.JSONEncoder, default=None
            Encoder used when writing rows as JSON objects to file during
            external merge sort.
        decoder: func, default=None
            Custom decoder function when reading archive rows from file. If not
            given, the default decoder will be used.
        """
        self.buffersize = buffersize if buffersize else config.SORTBUFFER()
        self.encoder = encoder
        self.decoder = decoder

    def _mergesort(
        self, buffer: List, filenames: List[str], columns: DocumentSchema,
        keys: List[int]
    ) -> str:
        """2-way merge sort for sorted blocks of an input document.

        Adopted from
        https://github.com/richardpenman/csvsort/blob/master/__init__.py

        Returns the name of the file containg the sorted output.

        Parameters
        ----------
        buffer: list
            Buffer of rows from the input file that were not written to file.
        filename: list
            List of names for sorted temporary files that were generated by the
            split method.
        columns: list of string
            List of column names.
        keys: list
            List of index positions for sort columns.

        Returns
        -------
        str
        """
        # Merge the buffer if it contains rows.
        if buffer:
            mergefile = filenames[0]
            with TempFile(delete=False, mode='w', newline='') as f_out:
                with JsonWriter(filename=f_out, encoder=self.encoder) as writer:
                    files = [
                        decorated_buffer(buffer, keys),
                        decorated_file(mergefile, keys, self.decoder)
                    ]
                    writer.write(columns)
                    for _, row in heapq.merge(*files):
                        writer.write(row)
                filenames = filenames[1:] + [f_out.name]
            os.remove(mergefile)
        while len(filenames) > 1:
            mergefiles, filenames = filenames[:2], filenames[2:]
            with TempFile(delete=False, mode='w', newline='') as f_out:
                with JsonWriter(filename=f_out, encoder=self.encoder) as writer:
                    files = [decorated_file(f, keys, self.decoder) for f in mergefiles]
                    writer.write(columns)
                    for _, row in heapq.merge(*files):
                        writer.write(row)
                filenames.append(f_out.name)
            for filename in mergefiles:
                os.remove(filename)
        return filenames[0]

    def sorted(self, doc: Document, keys: List[int]) -> Document:
        """Sort a input document based on the given columns.

        Returns a document that wraps a temporary file with the sorted
        document rows.

        Parameters
        ----------
        doc: histore.document.base.Document
            Input document.
        keys: list
            List of index positions for sort columns.

        Returns
        -------
        histore.document.base.Document
        """
        columns = doc.columns
        try:
            reader = doc.open()
            buffer, filenames = self._split(
                reader=reader,
                columns=columns,
                keys=keys
            )
        finally:
            doc.close()
        if not filenames:
            buffer.sort(key=lambda row: util.keyvalue(row[2], keys))
            return InMemoryDocument(columns=columns, rows=buffer)
        else:
            filename = self._mergesort(
                buffer=buffer,
                filenames=filenames,
                columns=columns,
                keys=keys
            )
            # Import JsonDocument here to avoid cyclic imports.
            from histore.document.json.base import JsonDocument
            return JsonDocument(
                filename=filename,
                encoder=self.encoder,
                decoder=self.decoder,
                delete_on_close=True
            )

    def _split(
        self, reader: DocumentIterator, columns: DocumentSchema,
        keys: List[int]
    ) -> Tuple[List, List[str]]:
        """Split a input document into blocks of maximum size.

        Individual blocks are written to temporary files on disk in JSON format.
        Only the final buffer is maintained in memory. Returns the memory buffer
        and the list of names for temporary files that are created.

        Adopted from
        https://github.com/richardpenman/csvsort/blob/master/__init__.py

        Parameters
        ----------
        reader: histore.document.base.DocumentIterator
            Iterator for rowid, cell values in an input document.
        columns: list of string
            List of column names.
        keys: list
            List of index positions for sort columns.

        Returns
        -------
        list(rows), list(filenames)
        """
        # Split document rows into blocks
        buffer = list()
        current_size = 0
        tmp_filenames = list()
        for row in reader:
            buffer.append(row)
            current_size += sys.getsizeof(row)
            if current_size > self.buffersize:
                # Sort buffer.
                buffer.sort(key=lambda row: util.keyvalue(row[2], keys))
                # Write buffer to disk.
                with TempFile(delete=False, mode='w', newline='') as f_out:
                    with JsonWriter(filename=f_out, encoder=self.encoder) as writer:
                        writer.write(columns)
                        for row in buffer:
                            writer.write(row)
                    # Append file name to result file name list.
                    tmp_filenames.append(f_out.name)
                # Clear buffer.
                buffer = list()
                current_size = 0
        return buffer, tmp_filenames


# -- Helper Functions ---------------------------------------------------------


def decorated_buffer(buffer: List, columns: List[int]):
    """Iterator for in-memory buffer of data rows.

    Parameters
    ----------
    buffer: list
        Buffer of rows from the input file that were not written to file.
    columns: list
        List of index positions for sort columns.
    """
    buffer.sort(key=lambda row: util.keyvalue(row[2], columns))
    for row in buffer:
        yield util.keyvalue(row[2], columns), row


def decorated_file(filename: str, columns: List[int], decoder: Callable):
    """Iterator for sorted file block. From
    https://github.com/richardpenman/csvsort/blob/master/__init__.py

    Parameters
    ----------
    filename: string
        Temporary file (block) on disk.
    columns: list
        List of index positions for sort columns.
    decoder: func
        Custom decoder function when reading archive rows from file. If not
        given, the default decoder will be used.
    """
    with JsonIterator(filename, decoder=decoder) as reader:
        for row in reader:
            yield util.keyvalue(row[2], columns), row



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2021 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Manager implementation for archives that are maintained in main memory."""

from typing import Callable, Dict, Optional, Union

from histore.archive.base import Archive, InputDocument, PrimaryKey
from histore.archive.manager.base import ArchiveManager
from histore.archive.manager.descriptor import ArchiveDescriptor
from histore.document.base import InputDescriptor

import histore.util as util


class VolatileArchiveManager(ArchiveManager):
    """The volatile archive manager maintains a set of archives in main memory.
    """
    def __init__(self):
        """Initialize the dictionaries that maintain archive descriptors and
        archive objects.
        """
        self._archives = dict()
        self._descriptors = dict()

    def archives(self) -> Dict[str, ArchiveDescriptor]:
        """Get dictionary of archive descriptors. The returned dictionary maps
        archive identifier to their descriptor.

        Returns
        -------
        dict(string: histore.archive.manager.descriptor.ArchiveDescriptor)
        """
        return self._descriptors

    def create(
        self, name: Optional[str] = None, description: Optional[str] = None,
        encoder: Optional[str] = None, decoder: Optional[str] = None,
        serializer: Union[Dict, Callable] = None, doc: Optional[InputDocument] = None,
        primary_key: Optional[PrimaryKey] = None, snapshot: Optional[InputDescriptor] = None,
        sorted: Optional[bool] = False, buffersize: Optional[float] = None,
        validate: Optional[bool] = False
    ) -> ArchiveDescriptor:
        """Create a new archive object under a given unique name.

        For archives that are keyed by a primary key, the input document for
        the first dataset snapshot  has to be provided. This snapshot will be
        loaded into the archive.

        Raises a ValueError if an archive with the given name exists.

        Parameters
        ----------
        name: string, default=None
            Descriptive name that is associated with the archive.
        description: string, default=None
            Optional long description that is associated with the archive.
        encoder: string, default=None
            Ignored. Full package path for the Json encoder class that is used
            by the persistent archive. Included for API completeness.
        decoder: string, default=None
            Ignored. Full package path for the Json decoder function that is
            used by the persistent archive. Included for API completeness.
        serializer: dict or callable, default=None
            Ignored. Dictionary or callable that returns a dictionary that
            contains the specification for the serializer. Included for API
            completeness.
        doc: histore.archive.base.InputDocument, default=None
            Input document representing the initial dataset snapshot that is
            being loaded into the archive.
        primary_key: string or list, default=None
            Column(s) that are used to generate identifier for snapshot rows.
        snapshot: histore.document.base.InputDescriptor, default=None
            Optional metadata for the created snapshot.
        sorted: bool, default=False
            Flag indicating if the document is sorted by the optional primary
            key attributes. Ignored if the archive is not keyed.
        buffersize: float, default=None
            Maximum size (in bytes) for the memory buffer when sorting the
            input document.
        validate: bool, default=False
            Validate that the resulting archive is in proper order before
            committing the action.

        Returns
        -------
        histore.archive.manager.descriptor.ArchiveDescriptor

        Raises
        ------
        ValueError
        """
        # Ensure that the archive name is unique.
        if self.get_by_name(name) is not None:
            raise ValueError("archive '{}' already exists".format(name))
        # Create an unique identifier for the new archive.
        identifier = util.get_unique_identifier()
        # Create in-memory archive. This will load the initial snapshot if a
        # document is given.
        archive = Archive(
            doc=doc,
            primary_key=primary_key,
            descriptor=snapshot,
            sorted=sorted,
            buffersize=buffersize,
            validate=validate
        )
        # Create the descriptor for the new archive.
        descriptor = ArchiveDescriptor.create(
            identifier=identifier,
            name=name,
            description=description,
            primary_key=archive.store.primary_key()
        )
        self._archives[identifier] = archive
        self._descriptors[identifier] = descriptor
        return descriptor

    def delete(self, identifier: str):
        """Delete the archive with the given identifier.

        Parameters
        ----------
        identifier: string
            Unique archive identifier
        """
        if self.contains(identifier):
            del self._archives[identifier]
            del self._descriptors[identifier]

    def get(self, identifier: str) -> Archive:
        """Get the archive that is associated with the given identifier. Raises
        a ValueError if the identifier is unknown.

        Parameters
        ----------
        identifier: string
            Unique archive identifier

        Returns
        -------
        histore.archive.base.Archive

        Raises
        ------
        ValueError
        """
        if identifier not in self._archives:
            raise ValueError("unknown archive '{}''".format(identifier))
        return self._archives[identifier]

    def rename(self, identifier: str, name: str):
        """Rename the specified archive. Raises a ValueError if the identifier
        is unknown or if an archive with the given name exist.

        Parameters
        ----------
        identifier: string
            Unique archive identifier
        name: string
            New archive name.

        Raises
        ------
        ValueError
        """
        archive = self._descriptors.get(identifier)
        if archive is None:
            raise ValueError("unknown archive '{}''".format(identifier))
        if archive.name() == name:
            # Do nothing if the archive name matches the new name.
            return
        # Raise an error if another archive with the new name exists.
        # Ensure that the archive name is unique.
        if self.get_by_name(name) is not None:
            raise ValueError("archive '{}' already exists".format(name))
        archive.rename(name)



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2021 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Helper methods to access configuration parameters. Configuration parameters
are maintained in environment variables.
"""

from appdirs import user_cache_dir

import os


"""Environment variables."""
# Base directory for all files that are created by histore components.
ENV_HISTORE_BASEDIR = 'HISTORE_BASEDIR'
# Database connection URL for archive manager.
ENV_HISTORE_DBCONNECT = 'HISTORE_DBCONNECT'
# Size of the sort buffer for CSV files that are sorted using external memory.
ENV_HISTORE_SORTBUFFER = 'HISTORE_SORTBUFFER'


def BASEDIR() -> str:
    """Get value for environment variable HISTORE_BASEDIR. The default value
    if the variable is not set is $HOME/.histore.

    Returns
    -------
    string
    """
    value = os.environ.get(ENV_HISTORE_BASEDIR)
    if value is None or value == '':  # pragma: no cover
        # Use the default value if the environment variable is not set.
        value = os.path.join(user_cache_dir(appname=__name__.split('.')[0]), '.histore')
    return value


def DBCONNECT() -> str:
    """Get value for environment variable HISTORE_DBCONNECT. The default value
    for the database connector is not defined (None).

    Returns
    -------
    string
    """
    return os.environ.get(ENV_HISTORE_DBCONNECT)


def SORTBUFFER() -> float:
    """Get value for environment variable HISTORE_SORTBUFFER. The default value
    if the variable is not set is 50% of the available main-memory.

    Returns
    -------
    float
    """
    value = os.environ.get(ENV_HISTORE_SORTBUFFER)
    if value is None or value == '':
        # Use 50% of the available main-memory as buffer size. Ensure that the
        # buffer is at least 1MB.
        import psutil
        mem = psutil.virtual_memory()
        value = max(mem.available / 2, 1024 * 1024)
    else:
        value = float(value)
    return value



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2021 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Implementation of the document interface for documents that have been read
completely into main memory.
"""

from __future__ import annotations
from typing import List, Optional, Tuple

import pandas as pd

from histore.document.base import DataRow, Document, DocumentIterator, RowIndex, document_to_df
from histore.document.schema import DocumentSchema

import histore.util as util


class InMemoryDocumentIterator(DocumentIterator):
    """Document iterator for in-memory documents."""
    def __init__(self, rows: List[Tuple[int, RowIndex, DataRow]]):
        """Initialize the list of document rows.

        Parameters
        ----------
        rows: list
            List of document rows. Each row in the list is a tuple of row
            position, row index, and cell values.
        """
        self.rows = rows
        self._readindex = 0

    def close(self):
        """Set list of rows to None."""
        self.rows = None

    def next(self) -> Tuple[int, RowIndex, DataRow]:
        """Read the next row in the document.

        Returns the row position, row index and the list of cell values for each
        of the document columns. Raises a StopIteration error if an attempt is
        made to read past the end of the document.

        Returns
        -------
        tuple of int, histore.document.base.RowIndex, histore.document.base.DataRow
        """
        try:
            rowpos, rowidx, values = self.rows[self._readindex]
        except (IndexError, TypeError):
            # If an attempt is made to read past the current row array length
            # either an IndexError is raised or a TypeError if the row array
            # has been set to None by the close method.
            raise StopIteration()
        self._readindex += 1
        return rowpos, rowidx, values


class InMemoryDocument(Document):
    """The in-memory document maintains a array of document rows (each as a
    tuple of row position, row index, and cell values). The row position is the
    original position of the row in a sorted document.
    """
    def __init__(self, columns: DocumentSchema, rows: List[Tuple[int, RowIndex, DataRow]]):
        """Initialize document schema and rows.

        Parameters
        ----------
        columns: list
            List of column names. The number of values in each row is expected
            to be the same as the number of columns and the order of values in
            each row is expected to correspond to their respective column in
            this list.
        rows: list
            List of document rows. Each row in the list is a tuple of row
            position, row index, and cell values.
        """
        super(InMemoryDocument, self).__init__(columns=columns)
        self.rows = rows

    def close(self):
        """Set list of rows to None when closing the document."""
        self.rows = None

    def open(self) -> InMemoryDocumentIterator:
        """Open the document to get a iterator for the rows in the document.

        Returns
        -------
        histore.document.base.DocumentIterator
        """
        return InMemoryDocumentIterator(rows=self.rows)

    def to_df(self) -> pd.DataFrame:
        """Create data frame from the document rows.

        Returns
        -------
        pd.DataFrame
        """
        return document_to_df(self)

    def sorted(self, keys: List[int], buffersize: Optional[float] = None) -> InMemoryDocument:
        """Sort the document rows based on the values in the key columns.

        Key columns are specified by their index position. Returns a new
        document.

        Parameters
        ----------
        keys: list of int
            Index position of sort columns.
        buffersize: float, default=None
            Maximum size (in bytes) of file blocks that are kept in main-memory.
            Ignored. Included for API completeness.

        Returns
        -------
        histore.document.mem.InMemoryDocument
        """
        return InMemoryDocument(
            columns=self.columns,
            rows=sorted(self.rows, key=lambda row: util.keyvalue(row[2], keys))
        )


class Schema(InMemoryDocument):
    """Empty document that only defines the document schema."""
    def __init__(self, columns: DocumentSchema):
        """Initialize document schema.

        Parameters
        ----------
        columns: list
            List of column names.
        """
        super(Schema, self).__init__(columns=columns, rows=[])



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2021 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Document wrapper around data frames."""

from __future__ import annotations
from typing import Any, List, Optional, Tuple

import math
import pandas as pd

from histore.document.base import DataRow, Document, DocumentIterator, RowIndex


class DataFrameIterator(DocumentIterator):
    """Document iterator for data frames."""
    def __init__(self, df: pd.DataFrame, readorder: Optional[List[int]] = None):
        """Initialize the data frame and the row read order.

        If no read order is given the read order is determined by the sorted
        data frame index.

        Parameters
        ----------
        df: pandas.DataFrame
            Pandas data frame that is being wrapped by the document class.
        readorder: list of int, default=None
            Order of row positions when iterating through the dataset.
        """
        self._df = df
        self.readorder = readorder if readorder is not None else rowindex_readorder(df)
        self._readindex = 0

    def close(self):
        """Set the associated data frame and read order to None."""
        self._df = None
        self.readorder = None

    def next(self) -> Tuple[int, RowIndex, DataRow]:
        """Read the next row in the document.

        Returns the row position, row index and the list of cell values for each
        of the document columns. Raises a StopIteration error if an attempt is
        made to read past the end of the document.

        Returns
        -------
        tuple of int, histore.document.base.RowIndex, histore.document.base.DataRow
        """
        rowpos = self._readindex
        try:
            origpos = self.readorder[rowpos]
        except IndexError:
            # Raise error if the end of the data frame has been reached.
            raise StopIteration()
        rowidx = self._df.index[origpos]
        # By definition, a row with an index that is None or has a negative
        # value will be treated as a new row.
        if rowidx is None or rowidx < 0:
            rowidx = -1
        values = [nan_to_none(v) for v in self._df.iloc[origpos]]
        self._readindex += 1
        return origpos, rowidx, values


class DataFrameDocument(Document):
    """Create an in-memory document for a pandas data frame."""
    def __init__(self, df: pd.DataFrame, readorder: Optional[List[int]] = None):
        """Initialize the data frame that represents the document.

        Parameters
        ----------
        df: pandas.DataFrame
            Pandas data frame that is being wrapped by the document class.
        readorder: list of int, default=None
            Order of row positions when iterating through the dataset.
        """
        super(DataFrameDocument, self).__init__(columns=list(df.columns))
        self._df = df
        self.readorder = readorder

    def close(self):
        """Set the internal reference to the data frame to None when the
        document is closed.
        """
        self._df = None

    def open(self) -> DataFrameIterator:
        """Open the document to get a iterator for the rows in the document.

        Returns
        -------
        histore.document.df.DataFrameIterator
        """
        return DataFrameIterator(df=self._df, readorder=self.readorder)

    def to_df(self) -> pd.DataFrame:
        """Return the associated data frame.

        Returns
        -------
        pd.DataFrame
        """
        return self._df

    def sorted(self, keys: List[int], buffersize: Optional[float] = None) -> DataFrameDocument:
        """Sort the document rows based on the values in the key columns.

        Key columns are specified by their index position. Returns a new
        document.

        Parameters
        ----------
        keys: list of int
            Index position of sort columns.
        buffersize: float, default=None
            Maximum size (in bytes) of file blocks that are kept in main-memory.
            Ignored. Included for API completeness.

        Returns
        -------
        histore.document.df.DataFrameDocument
        """
        rows = list()
        if len(keys) == 1:
            ridx = keys[0]
            for row in self._df.itertuples(index=False, name=None):
                rows.append((len(rows), row[ridx]))
        else:
            for row in self._df.itertuples(index=False, name=None):
                rows.append((len(rows), tuple(row[k] for k in keys)))
        return DataFrameDocument(
            df=self._df,
            readorder=[pos for pos, _ in sorted(rows, key=lambda x: x[1])]
        )


# -- Helper Functions ---------------------------------------------------------

def nan_to_none(value: Any) -> Any:
    """Convert NaN values to None.

    Parameters
    ----------
    value: any
        Cell value.

    Returns
    -------
    any
    """
    if isinstance(value, int) or isinstance(value, float):
        return None if math.isnan(value) else value
    return value


def rowindex_readorder(df: pd.DataFrame) -> List[int]:
    """Get read order for rows in a pandas data frame based on their index.

    Sorts the rows based on the index and returns a list with initial row
    positions in sorted order. Index values that are None or -1 are considered
    new rows that are moved to the end of the sorted list.

    Parameters
    ----------
    df: pd.DataFrame
        Input data frame.

    Returns
    -------
    list of int
    """
    keyed_rows, new_rows = list(), list()
    for pos, index in zip(range(len(df.index)), df.index):
        if index is None or index == -1:
            new_rows.append(pos)
        else:
            keyed_rows.append((pos, index))
    return [pos for pos, _ in sorted(keyed_rows, key=lambda x: x[1])] + new_rows



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2021 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Writer for CSV documents."""

import csv
import gzip

from typing import List, Optional

from histore.document.base import Scalar


class CSVWriter(object):
    """Context manager for an open CSV file writer."""
    def __init__(
        self, filename: str, delim: Optional[str] = None, compressed: Optional[bool] = None,
        encoding: Optional[str] = None, none_as: Optional[str] = None
    ):
        """Initialize the CSV writer.

        Parameters
        ----------
        filename: string
            Path to the output file.
        delim: string, default=None
            The column delimiter used in the CSV file.
        compressed: bool, default=None
            Flag indicating if the file contents are to be compressed using
            gzip.
        encoding: string, default=None
            The csv file encoding e.g. utf-8, utf-16 etc.
        none_as: string, default=None
            String that is used to encode None values in the output file. If
            given, all cell values that are None are substituted by the string.
        """
        # Open file depending on whether it is gzip compressed or not.
        if compressed:
            self.file = gzip.open(filename, 'wt', newline='')
        else:
            self.file = open(filename, 'w', newline='', encoding=encoding)
        # Open te CSV writer.
        self.writer = csv.writer(self.file, delimiter=delim if delim is not None else ',')
        self.none_as = none_as

    def __enter__(self):
        """Enter method for the context manager."""
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Close the associated file handle when the context manager exits."""
        self.close()
        return False

    def close(self):
        """Close the associated file handle and set it to None (to avoid
        repeated attempts to close an already closed file).
        """
        if self.file is not None:
            self.file.close()
            self.file = None

    def write(self, row: List[Scalar]):
        """Write a row to the CSV file.

        Parameters
        ----------
        row: list
            List of cell values for a dataset row.
        """
        if self.none_as is not None:
            row = [v if v is not None else self.none_as for v in row]
        return self.writer.writerow(row)



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2021 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Implementation of the archive reader for the in-memory archive."""

from histore.archive.reader import ArchiveReader


class BufferedReader(ArchiveReader):
    """Reader for a list of archive rows that are kept in main memory."""
    def __init__(self, rows):
        """Initialize the list of rows that the reader will return.

        Parameters
        ----------
        rows: list(histore.archive.row.ArchiveRow)
            List of rows in the input stream.
        """
        self.rows = rows
        # Maintain a read index that points to the next row in the input
        # stream.
        self.read_index = 0

    def close(self):
        """Set row buffer to None."""
        self.rows = None

    def next(self):
        """Read the next row in the dataset archive. Returns None if the end of
        the archive rows has been reached.

        Returns
        -------
        histore.archive.row.ArchiveRow
        """
        # Get the next row that the read index points to. Advance the read
        # index before returning that row.
        try:
            row = self.rows[self.read_index]
            self.read_index += 1
            return row
        except (IndexError, TypeError):
            self.close()
            return None



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2021 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Abstract class for managers that maintain a set of archives."""

from abc import ABCMeta, abstractmethod
from typing import Callable, Dict, List, Optional, Union

from histore.archive.base import Archive, InputDocument, PrimaryKey
from histore.archive.manager.descriptor import ArchiveDescriptor
from histore.document.base import InputDescriptor


class ArchiveManager(metaclass=ABCMeta):
    """The manager for archives implements a factory pattern for creating and
    accessing different types of archives. The type of archive (e.g. persistent
    or volatile is implementaiton dependent. Matdata about each archive is
    maintained in an archive descriptor.
    """
    @abstractmethod
    def archives(self) -> Dict[str, ArchiveDescriptor]:
        """Get dictionary of archive descriptors. The returned dictionary maps
        archive identifier to their descriptor.

        Returns
        -------
        dict(string: histore.archive.manager.descriptor.ArchiveDescriptor)
        """
        raise NotImplementedError()  # pragma: no cover

    def contains(self, identifier: str) -> bool:
        """Returns True if an archive with the given identifier exists.

        Parameters
        ----------
        identifier: string
            Unique archive identifier

        Returns
        -------
        bool
        """
        return identifier in self.archives()

    @abstractmethod
    def create(
        self, name: Optional[str] = None, description: Optional[str] = None,
        encoder: Optional[str] = None, decoder: Optional[str] = None,
        serializer: Union[Dict, Callable] = None, doc: Optional[InputDocument] = None,
        primary_key: Optional[PrimaryKey] = None, snapshot: Optional[InputDescriptor] = None,
        sorted: Optional[bool] = False, buffersize: Optional[float] = None,
        validate: Optional[bool] = False
    ) -> ArchiveDescriptor:
        """Create a new archive object under a given unique name.

        For archives that are keyed by a primary key, the input document for
        the first dataset snapshot  has to be provided. This snapshot will be
        loaded into the archive.

        Raises a ValueError if an archive with the given name exists.

        Parameters
        ----------
        name: string, default=None
            Descriptive name that is associated with the archive.
        description: string, default=None
            Optional long description that is associated with the archive.
        encoder: string, default=None
            Full package path for the Json encoder class that is used by the
            persistent archive.
        decoder: string, default=None
            Full package path for the Json decoder function that is used by the
            persistent archive.
        serializer: dict or callable, default=None
            Dictionary or callable that returns a dictionary that contains the
            specification for the serializer. The serializer specification is
            a dictionary with the following elements:
            - ``clspath``: Full package target path for the serializer class
            that is instantiated.
            - ``kwargs`` : Additional arguments that are passed to the
            constructor of the created serializer instance.
            Only ``clspath`` is required.
        doc: histore.archive.base.InputDocument, default=None
            Input document representing the initial dataset snapshot that is
            being loaded into the archive.
        primary_key: string or list, default=None
            Column(s) that are used to generate identifier for snapshot rows.
        snapshot: histore.document.base.InputDescriptor, default=None
            Optional metadata for the created snapshot.
        sorted: bool, default=False
            Flag indicating if the document is sorted by the optional primary
            key attributes. Ignored if the archive is not keyed.
        buffersize: float, default=None
            Maximum size (in bytes) for the memory buffer when sorting the
            input document.
        validate: bool, default=False
            Validate that the resulting archive is in proper order before
            committing the action.

        Returns
        -------
        histore.archive.manager.descriptor.ArchiveDescriptor

        Raises
        ------
        ValueError
        """
        raise NotImplementedError()  # pragma: no cover

    @abstractmethod
    def delete(self, identifier: str):
        """Delete the archive with the given identifier.

        Parameters
        ----------
        identifier: string
            Unique archive identifier
        """
        raise NotImplementedError()  # pragma: no cover

    @abstractmethod
    def get(self, identifier: str) -> Archive:
        """Get the archive that is associated with the given identifier. Raises
        a ValueError if the identifier is unknown.

        Parameters
        ----------
        identifier: string
            Unique archive identifier

        Returns
        -------
        histore.archive.base.Archive

        Raises
        ------
        ValueError
        """
        raise NotImplementedError()  # pragma: no cover

    def get_by_name(self, name: str) -> ArchiveDescriptor:
        """Get descriptor for the archive with the given name. If no archive
        with that name exists None is returned.

        Parameters
        ----------
        name: string
            Archive name

        Returns
        -------
        histore.archive.manager.descriptor.ArchiveDescriptor
        """
        for archive in self.archives().values():
            if archive.name() == name:
                return archive
        return None

    def list(self) -> List[ArchiveDescriptor]:
        """Get the list of descriptors for the maintained archives.

        Returns
        -------
        list(histore.archive.manager.descriptor.ArchiveDescriptor)
        """
        return list(self.archives().values())

    @abstractmethod
    def rename(self, identifier: str, name: str):
        """Rename the specified archive. Raises a ValueError if the identifier
        is unknown or if an archive with the given name exist.

        Parameters
        ----------
        identifier: string
            Unique archive identifier
        name: string
            New archive name.

        Raises
        ------
        ValueError
        """
        raise NotImplementedError()  # pragma: no cover



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2021 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Wrapper for database connections. The wrapper is used to open database
sessions as well as to create a fresh database.
"""

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker, scoped_session
from sqlalchemy.orm.session import Session

from typing import Optional

from histore.archive.manager.db.model import Base


"""Database connection Url for test purposes."""
TEST_URL = 'sqlite:///:memory:'


class DB(object):
    """Wrapper to establish a database connection and create the database
    schema.
    """
    def __init__(
        self, connect_url: str, web_app: Optional[bool] = False,
        echo: Optional[bool] = False
    ):
        """Initialize the database connection string.
        Parameters
        ----------
        connect_url: string
            Database connection string.
        web_app: bool, default=False
            Use scoped sessions for web applications if set to True.
        echo: bool, default=False
            Flag that controlls whether the created engine is verbose or not.

        Raises
        ------
        flowserv.error.MissingConfigurationError
        """
        self.web_app = web_app
        # Ensure that the connection URL is set.
        if echo:
            import logging
            logging.info('Connect to database Url %s' % (connect_url))
        self._engine = create_engine(connect_url, echo=echo)
        if web_app:
            self._session = scoped_session(sessionmaker(bind=self._engine))
        else:
            self._session = sessionmaker(bind=self._engine)

    def close(self):
        """Remove sessions from the scoped session maker if in web application
        mode.
        """
        if self.web_app:
            self._session.remove()

    def init(self):
        """Create all tables in the database model schema."""
        # Add import for modules that contain ORM definitions.
        import histore.archive.manager.db.model  # noqa: F401
        # Drop all tables first before creating them
        Base.metadata.drop_all(self._engine)
        Base.metadata.create_all(self._engine)

    def session(self):
        """Create a new database session instance. The sessoin is wrapped by a
        context manager to properly manage the session scope.

        Returns
        -------
        flowserv.model.database.SessionScope
        """
        return SessionScope(self._session())


class SessionScope(object):
    """Context manager for providing transactional scope around a series of
    database operations.
    """
    def __init__(self, session: Session):
        """Initialize the database session.

        Parameters
        ----------
        session: sqlalchemy.orm.session.Session
            Database session.
        """
        self.session = session

    def __enter__(self):
        """Return the managed database session object.

        Returns
        -------
        sqlalchemy.orm.session.Session
        """
        return self.session

    def __exit__(self, exc_type, exc_value, exc_traceback):
        """Commit or rollback transaction depending on the exception type.
        Does not surpress any exceptions.
        """
        if exc_type is None:
            try:
                self.session.commit()
            except Exception:
                self.session.rollback()
                raise
            finally:
                self.session.close()
        else:
            try:
                self.session.rollback()
            finally:
                self.session.close()



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2021 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Command line interface to interact with a manager for archives on the local
file system.
"""

import click
import os
import sys

from histore.archive.manager.persist import PersistentArchiveManager
from histore.cli.archive import create_archive, delete_archive, list_archives, rename_archive
from histore.cli.snapshot import checkout_snapshot, commit_snapshot, list_snapshots

import histore.config as config


# -- Init the archive manager -------------------------------------------------

@click.command(name='init')
@click.option(
    '-b', '--basedir',
    required=False,
    type=click.Path(file_okay=False, dir_okay=True),
    help='Base directory for archive files'
)
@click.option(
    '-c', '--dbconnect',
    required=False,
    help='Connect URL for the database'
)
def init_manager(basedir, dbconnect):
    """Initialize the archive manager directory."""
    # Test if the base directory exists and is empty.
    basedir = basedir if basedir is not None else config.BASEDIR()
    if os.path.isdir(basedir):
        if os.listdir(basedir):
            click.echo('Not an empty directory {}.'.format(basedir))
            sys.exit(-1)
    # Create instance of persistent archive manager to setup directories and
    # files.
    PersistentArchiveManager(
        basedir=basedir,
        dbconnect=dbconnect,
        create=True
    )
    click.echo("Initialized in {}.".format(os.path.abspath(basedir)))


# -- Create command group -----------------------------------------------------

@click.group()
def cli():  # pragma: no cover
    """Command line interface for HISTORE archive manager."""
    pass


cli.add_command(init_manager)
cli.add_command(checkout_snapshot)
cli.add_command(commit_snapshot)
cli.add_command(create_archive)
cli.add_command(delete_archive)
cli.add_command(list_archives)
cli.add_command(list_snapshots)
cli.add_command(rename_archive)



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2021 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Collection of helper classes to read data frames from CSV files."""

from typing import List, Optional, Tuple

import os

from histore.document.base import DataRow, Document, DocumentIterator, RowIndex
from histore.document.csv.reader import CSVReader
from histore.document.csv.writer import CSVWriter
from histore.document.external import ExternalDocument
from histore.document.schema import DocumentSchema
from histore.document.sort import SortEngine


class CSVIterator(DocumentIterator):
    """Document iterator for CSV documents."""
    def __init__(
        self, filename: str, delim: Optional[str] = None, compressed: Optional[bool] = None,
        quotechar: Optional[str] = '"', encoding: Optional[str] = None,
        none_is: Optional[str] = None, skip_header: Optional[bool] = False
    ):
        """Initialize the input file.

        Parameters
        ----------
        filename: string
            Path to the input file.
        delim: string, default=None
            The column delimiter used in the CSV file.
        compressed: bool, default=None
            Flag indicating if the file contents are compressed using gzip.
        quotechar: string, default='"'
            CSV quote char.
        encoding: string, default=None
            The csv file encoding e.g. utf-8, utf-16 etc.
        none_is: string, default=None
            String that was used to encode None values in the input file. If
            given, all cell values that match the given string are substituted
            by None.
        skip_header: bool, default=False
            Skip first row if the CSV file contains header information.
        """
        self.reader = CSVReader(
            filename=filename,
            delim=delim,
            compressed=compressed,
            quotechar=quotechar,
            encoding=encoding,
            none_is=none_is
        )
        # Skip the column name row.
        if skip_header:
            next(self.reader)
        # Initialize the row index.
        self._rowindex = 0

    def close(self):
        """Close the associated Json reader."""
        self.reader.close()

    def next(self) -> Tuple[int, RowIndex, DataRow]:
        """Read the next row in the document.

        Returns the row position, row index and the list of cell values for each
        of the document columns. Raises a StopIteration error if an attempt is
        made to read past the end of the document.

        Returns
        -------
        tuple of int, histore.document.base.RowIndex, histore.document.base.DataRow
        """
        # Get the row values.
        values = next(self.reader)
        # Increment the row counter.
        rowidx = self._rowindex
        self._rowindex += 1
        # Return row information.
        return rowidx, rowidx, values


class CSVFile(ExternalDocument):
    """Read and write documents that are stored as CSV files."""
    def __init__(
        self, filename: str, header: Optional[DocumentSchema] = None,
        delim: Optional[str] = None, compressed: Optional[bool] = None,
        quotechar: Optional[str] = '"', encoding: Optional[str] = None,
        none_is: Optional[str] = None, skip_lines: Optional[int] = None,
        delete_on_close: Optional[bool] = False
    ):
        """Set the file name, delimiter and the flag that indicates if the file
        is compressed using gzip.

        If the file is opened for writing and no header is given no attempt is
        made to reader the header from file.

        Parameters
        ----------
        filename: string
            Path to the CSV file that is being read.
        header: list of string, default=None
            Optional header. If no header is given it is assumed that the first
            row in the CSV file contains the header information.
        delim: string, default=None
            The column delimiter used in the CSV file.
        compressed: bool, default=None
            Flag indicating if the file contents have been compressed using
            gzip.
        encoding: string, default=None
            The csv file encoding e.g. utf-8, utf-16 etc.
        none_is: string, default=None
            String that was used to encode None values in the input file. If
            given, all cell values that match the given string are substituted
            by None.
        skip_lines: int, default=None
            Skip the first n lines.
        delete_on_close: bool, default=False
            If True, delete the file when the document is closed.
        """
        self.filename = filename
        self.quotechar = quotechar
        self.none_is = none_is
        self.skip_lines = skip_lines
        # Keep track if the header information was None intially, i.e., if the
        # header is in the file or not.
        self._has_header = header is None
        # Infer the delimiter from the filename if not given. Files that have
        # '.tsv' in their name are expscted to be tab-dlimited. All other files
        # are expected to be standard CSV files. In the future we may want to
        # use a sniffer here.
        if delim is None:
            if filename.endswith('.tsv') or filename.endswith('.tsv.gz'):
                delim = '\t'
            else:
                delim = ','
        self.delim = delim
        # Infer compression if gzip is not set. Files that end with '.gz' are
        # assumed to be compressed using gzip.
        if compressed is None:
            compressed = filename.endswith('.gz')
        self.compressed = compressed
        self.encoding = encoding
        self.none_is = none_is
        # Read the header information from the first line of the input file if
        # no header is given.
        if header is None and os.path.isfile(filename):
            # Open the reader for the input file to get the list of column names
            # from the first row. If the file is empty the list of rows is empty.
            reader = CSVReader(
                filename=self.filename,
                delim=self.delim,
                compressed=self.compressed,
                quotechar=self.quotechar,
                encoding=self.encoding,
                none_is=self.none_is
            )
            header = next(reader)
            reader.close()
        super(CSVFile, self).__init__(
            columns=header,
            filename=filename,
            delete_on_close=delete_on_close
        )

    def open(self) -> CSVIterator:
        """Get a row iterator for the associated CSV file.

        Returns
        -------
        histore.document.csv.base.CSVIterator
        """
        return CSVIterator(
            filename=self.filename,
            delim=self.delim,
            compressed=self.compressed,
            quotechar=self.quotechar,
            encoding=self.encoding,
            none_is=self.none_is,
            skip_header=self._has_header
        )

    def sorted(self, keys: List[int], buffersize: Optional[float] = None) -> Document:
        """Sort the document rows based on the values in the key columns.

        Key columns are specified by their index position. Returns a new
        document.

        Parameters
        ----------
        keys: list of int
            Index position of sort columns.
        buffersize: float, default=None
            Maximum size (in bytes) of file blocks that are kept in main-memory.

        Returns
        -------
        histore.document.base.Document
        """
        return SortEngine(buffersize=buffersize).sorted(doc=self, keys=keys)

    def writer(self) -> CSVWriter:
        """Get a CSV writer for the associated CSV file.

        Returns
        -------
        histore.document.csv.base.CSVWriter
        """
        return CSVWriter(
            filename=self.filename,
            delim=self.delim,
            compressed=self.compressed,
            encoding=self.encoding,
            none_as=self.none_is
        )



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2021 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Helper methods for timestamp management."""

from abc import ABCMeta, abstractmethod
from datetime import datetime
from dateutil.parser import isoparse
from dateutil.tz import UTC
from importlib import import_module
from typing import IO, Iterator, List, Optional, Union

import errno
import gzip
import os
import shutil
import uuid


# -- CLI ----------------------------------------------------------------------

def get_delimiter(delimiter):
    """Get delimiter. Replace encodings for tabulator with tab character.

    Parameters
    ----------
    delimiter: string
        One-character used to separate fields.

    Returns
    -------
    string
    """
    if delimiter is None:
        return ','
    if delimiter.lower() in ['tab', '\\t']:
        return '\t'
    return delimiter


# -- Datetime -----------------------------------------------------------------

def current_time() -> str:
    """Get ISO format string for the current time.

    Returns
    -------
    str
    """
    return utc_now().isoformat()


def to_datetime(timestamp: str) -> datetime:
    """Converts a timestamp string in ISO format into a datatime object in
    UTC timezone.

    Parameters
    ----------
    timstamp : string
        Timestamp in ISO format

    Returns
    -------
    datetime.datetime
        Datetime object
    """
    if isinstance(timestamp, datetime):
        return timestamp
    try:
        dt = isoparse(timestamp)
        if not dt.tzinfo == UTC:
            dt = dt.astimezone(UTC)
        return dt
    except ValueError:
        raise ValueError('unknown time format {}'.format(timestamp))


def to_localtime(ts):
    """Conbert timestamp in UTC timezone to local time.

    Parameters
    ----------
     ts: datetime.datetime
        Timestamp in UTC format

    Returns
    -------
    datetime.datetime
    """
    return ts.astimezone()


def utc_now():
    """Get the current time in UTC timezone.

    Returns
    -------
    datetime.datetime
    """
    return datetime.now(UTC)


# -- Dynamic import ----------------------------------------------------------

def import_obj(import_path):
    """Import an object (function or class) from a given package path.

    Parameters
    ----------
    import_path: string
        Full package target path for the imported object. Assumes that path
        components are separated by '.'.

    Returns
    -------
    any
    """
    pos = import_path.rfind('.')
    module_name = import_path[:pos]
    class_name = import_path[pos + 1:]
    module = import_module(module_name)
    return getattr(module, class_name)


# -- I/O ----------------------------------------------------------------------

class IOStream(metaclass=ABCMeta):
    def __init__(self, f):
        """Initialize the IO stream.

        Parameters
        ----------
        f: FileObject
            Open file object.
        """
        self.f = f

    def close(self):
        """Close the IO Stream."""
        self.f.close()

    @abstractmethod
    def write(self, line):  # pragma: no cover
        """Write sting line to output file.

        Parameters
        ----------
        line: string
            Output line that is bein written.
        """
        raise NotImplementedError()


class GZipFile(IOStream):
    """IO stream wrapper for gzip compressed files."""
    def __init__(self, f):
        """Initialize the IO stream.

        Parameters
        ----------
        f: FileObject
            Open file object.
        """
        super(GZipFile, self).__init__(f)

    def write(self, line):
        """Write sting line to output file.

        Parameters
        ----------
        line: string
            Output line that is bein written.
        """
        self.f.write(str.encode(line, 'utf8'))


class PlainTextFile(IOStream):
    """IO stream for plain text files."""
    def __init__(self, f):
        """Initialize the IO stream.

        Parameters
        ----------
        f: FileObject
            Open file object.
        """
        super(PlainTextFile, self).__init__(f)

    def write(self, line):
        """Write sting line to output file.

        Parameters
        ----------
        line: string
            Output line that is bein written.
        """
        self.f.write(line)


def cleardir(directory):
    """Remove all files in the given directory.

    Parameters
    ----------
    directory: string
        Path to directory that is being created.
    """
    for filename in os.listdir(directory):
        file = os.path.join(directory, filename)
        if os.path.isfile(file) or os.path.islink(file):
            os.unlink(file)
        else:
            shutil.rmtree(file)


def createdir(directory, abs=False):
    """Safely create the given directory path if it does not exist.

    Parameters
    ----------
    directory: string
        Path to directory that is being created.
    abs: boolean, optional
        Return absolute path if true

    Returns
    -------
    string
    """
    # Based on https://stackoverflow.com/questions/273192/
    if not os.path.exists(directory):
        try:
            os.makedirs(directory)
        except OSError as e:  # pragma: no cover
            if e.errno != errno.EEXIST:
                raise
    if abs:
        return os.path.abspath(directory)
    else:
        return directory


def inputstream(filename: Union[str, IO], compression: Optional[str] = None) -> Iterator[str]:
    """Open the given file for input. The compression mode string determines
    which compression algorithm is being used (or no compression if None).

    Parameters
    ----------
    compression: string, default=None
        String representing the compression mode for the output file.

    Returns
    -------
    histore.util.IOStream
    """
    if compression is None:
        f = open(filename, 'r') if isinstance(filename, str) else filename

        def file_iterator(f: IO):
            """Turn an open file handle into an iterator."""
            for line in f:
                yield line

        return file_iterator(f)
    elif compression == 'gzip':
        f = gzip.open(filename, 'rb')

        def file_iterator(f: IO):
            """Turn an open file handle into an iterator."""
            for line in f:
                yield line.decode('utf8')

        return file_iterator(f)

    raise ValueError('unknown compression mode {}'.format(compression))


def outputstream(filename, compression: Optional[str] = None) -> IOStream:
    """Open the given file for output. The compression mode string determines
    which compression algorithm is being used (or no compression if None).

    Parameters
    ----------
    compression: string, default=None
        String representing the compression mode for the output file.

    Returns
    -------
    histore.util.IOStream
    """
    if compression is None:
        return PlainTextFile(open(filename, 'w') if isinstance(filename, str) else filename)
    elif compression == 'gzip':
        return GZipFile(gzip.open(filename, 'wb'))
    raise ValueError('unknown compression mode {}'.format(compression))


# -- Sorting ------------------------------------------------------------------

def keyvalue(row: List, columns: List[int]):
    """Get the sort key for a given row. From
    https://github.com/richardpenman/csvsort/blob/master/__init__.py

    Parameters
    ----------
    row: list
        List of cell values in a document row.
    columns: list
        List of index positions for sort columns.

    Returns
    -------
    list
    """
    return [row[column] for column in columns]


# -- Unique identifier --------------------------------------------------------

def get_unique_identifier():
    """Create a new unique identifier.

    Returns
    -------
    string
    """
    return str(uuid.uuid4()).replace('-', '')



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2021 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Command line interface to interact with a manager for archives on the local
file system.
"""

from typing import Optional

import click
import csv
import pandas as pd
import sys

from histore.archive.manager.base import ArchiveManager
from histore.archive.manager.persist import PersistentArchiveManager

import histore.config as config
import histore.util as util


"""Datetime format string."""
DTF = '%Y-%m-%d %H:%M:%S'


# -- Create a new archive -----------------------------------------------------

@click.command(name='create')
@click.option(
    '-b', '--basedir',
    required=False,
    type=click.Path(file_okay=False, dir_okay=True),
    help='Base directory for archive files'
)
@click.option(
    '-c', '--dbconnect',
    required=False,
    help='Connect URL for the database'
)
@click.option(
    '-k', '--pk',
    required=False,
    type=str,
    help='Comma-separate list of primary key columns'
)
@click.option(
    '-d', '--delimiter',
    required=False,
    type=str,
    help='One-character string used to separate fields'
)
@click.option(
    '-q', '--quotechar',
    required=False,
    type=str,
    help='One-character string used to quote fields with special characters'
)
@click.option(
    '-z', '--gzip',
    is_flag=True,
    default=False,
    help='Gzip compressed'
)
@click.option(
    '-f', '--filename',
    type=click.Path(file_okay=True, dir_okay=False, exists=True)
)
@click.option(
    '-t', '--comment',
    required=False,
    type=str,
    help='Optional archive description'
)
@click.option(
    '-e', '--encoder',
    required=False,
    type=str,
    help='JSONEncoder class for the new archive'
)
@click.option(
    '-x', '--decoder',
    required=False,
    type=str,
    help='JSON decoder function for the new archive'
)
@click.argument('name')
def create_archive(
    basedir, dbconnect, pk, delimiter, quotechar, gzip, filename, comment,
    encoder, decoder, name
):
    """Create a new archive."""
    manager = get_manager(basedir, dbconnect=dbconnect)
    # Split primary key if it contains ','.
    primary_key = pk.split(',') if pk is not None else None
    if filename:
        # Read the data frame.
        doc = pd.read_csv(
            filename,
            delimiter=util.get_delimiter(delimiter),
            quotechar=quotechar if quotechar is not None else '"',
            quoting=csv.QUOTE_MINIMAL,
            compression='gzip' if gzip else None
        )
    else:
        doc = None
    try:
        manager.create(
            name=name,
            description=comment,
            primary_key=primary_key,
            doc=doc,
            encoder=encoder,
            decoder=decoder
        )
        click.echo('Archive created!')
    except ValueError as ex:
        click.echo('{}'.format(ex))
        sys.exit(-1)


# -- Delete archive -----------------------------------------------------------

@click.command(name='delete')
@click.option(
    '-b', '--basedir',
    required=False,
    type=click.Path(file_okay=False, dir_okay=True),
    help='Base directory for archive files'
)
@click.option(
    '-c', '--dbconnect',
    required=False,
    help='Connect URL for the database'
)
@click.option(
    '-f', '--force',
    is_flag=True,
    default=False,
    help='Delete without confirmation'
)
@click.argument('name')
def delete_archive(basedir, dbconnect, force, name):
    """Delete existing archive."""
    manager = get_manager(basedir, dbconnect=dbconnect)
    for archive in manager.list():
        if archive.name() == name:
            if not force:
                msg = "This will remove archive '{}' permanently"
                click.echo(msg.format(name))
                click.confirm('Continue?', default=True, abort=True)
            manager.delete(archive.identifier())
            click.echo("Archive '{}' deleted!".format(name))
            return
    click.echo("Unknown archive '{}'.".format(name))
    sys.exit(-1)


# -- List archives ------------------------------------------------------------

@click.command(name='list')
@click.option(
    '-b', '--basedir',
    required=False,
    type=click.Path(file_okay=False, dir_okay=True),
    help='Base directory for archive files'
)
@click.option(
    '-c', '--dbconnect',
    required=False,
    help='Connect URL for the database'
)
@click.option(
    '-d', '--bydate',
    is_flag=True,
    default=False,
    help='Sort by creation date'
)
def list_archives(basedir, dbconnect, bydate):
    """List names of existing archives."""
    manager = get_manager(basedir, dbconnect=dbconnect)
    archives = manager.list()
    if bydate:
        archives = sorted(archives, key=lambda a: a.created_at())
    else:
        archives = sorted(archives, key=lambda a: a.name())
    click.echo()
    click.echo('Archives')
    click.echo('--------')
    for archive in archives:
        ts = util.to_localtime(archive.created_at())
        click.echo('{} (created at {})'.format(
            archive.name(),
            ts.strftime('%Y-%m-%d %H:%M:%S')
        ))
    click.echo()


# -- Rename archive -----------------------------------------------------------

@click.command(name='rename')
@click.option(
    '-b', '--basedir',
    required=False,
    type=click.Path(file_okay=False, dir_okay=True),
    help='Base directory for archive files'
)
@click.option(
    '-c', '--dbconnect',
    required=False,
    help='Connect URL for the database'
)
@click.argument('oldname')
@click.argument('newname')
def rename_archive(basedir, dbconnect, oldname, newname):
    """Rename existing archive."""
    manager = get_manager(basedir, dbconnect=dbconnect)
    # Get archive with the old name
    archive = manager.get_by_name(oldname)
    if archive is None:
        click.echo("Unknown archive '{}'".format(oldname))
        sys.exit(-1)
    try:
        manager.rename(archive.identifier(), newname)
    except ValueError as ex:
        click.echo('{}'.format(ex))
        sys.exit(-1)


# -- Helper Functions ---------------------------------------------------------

def get_manager(
    basedir: str, dbconnect: Optional[str] = None
) -> ArchiveManager:
    """Create instance of persistent archive manager assuming that it has been
    initialized before.

    Parameters
    ----------
    basedir: string
        Base directory for the archive manager.
    dbconnect: string, default=None
        Database connect string.

    Returns
    -------
    histore.archive.manager.base.ArchiveManager
    """
    basedir = basedir if basedir is not None else config.BASEDIR()
    return PersistentArchiveManager(
        basedir=basedir,
        dbconnect=dbconnect,
        create=False
    )



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2021 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Nested-merge logic that merges rows in a snapshot document into an existing
dataset archive.
"""


def merge_rows(arch_reader, doc_reader, version, writer):
    """Merge rows in the given archive and database snapshot. Outputs the
    merged rows in the resulting archive to the given archive writer.

    Parameters
    ----------
    archive: histore.archive.reader.ArchiveReader
        Reader for archive containing previous snapshots of the dataset.
    doc_reader: histore.document.reader.DocumentReader
        Reader for rows in the merged dataset snapshot.
    version: int
        Identifier of the new archive version.
    writer: histore.archive.writer.ArchiveWriter
        Consumer for rows in the new archive version.
    """
    # Get the first row for each reader.
    arch_row = arch_reader.next()
    doc_row = doc_reader.next()
    while arch_row is not None and doc_row is not None:
        comp = arch_row.comp(doc_row.key)
        if comp < 0:
            # The row is not present in the document. Include the row without
            # changes to the new archive version and read the next archive row.
            writer.write_archive_row(arch_row)
            arch_row = arch_reader.next()
        elif comp > 0:
            # The document row is a new row. Create an archive row from the
            # document row and pass it to the writer. Progress the document
            # reader.
            writer.write_document_row(row=doc_row, version=version)
            doc_row = doc_reader.next()
        else:
            # Merge the archive row and the document row.
            arch_row = arch_row.merge(
                values=doc_row.values,
                pos=doc_row.pos,
                version=version
            )
            # Add the merged row to the new archive version and progress
            # both readers.
            writer.write_archive_row(arch_row)
            arch_row = arch_reader.next()
            doc_row = doc_reader.next()
    # Add remaining rows to the archive. Only one of the two conditions can
    # be true; either there are additional archive rows or there are additional
    # document rows.
    # Add remaining archive rows to the new archive version.
    while arch_row is not None:
        writer.write_archive_row(arch_row)
        arch_row = arch_reader.next()
    # Add remaining document rows to the new archive version.
    while doc_row is not None:
        # Output an archive row created from the document row.
        writer.write_document_row(row=doc_row, version=version)
        doc_row = doc_reader.next()
    doc_reader.close()



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2021 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

from sqlalchemy import Integer, String, Text
from sqlalchemy import Column, ForeignKey
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship
from sqlalchemy.types import TypeDecorator, Unicode

import json

from histore.archive.manager.descriptor import ArchiveDescriptor

import histore.util as util


# -- ORM Model ----------------------------------------------------------------

"""Base class for all database tables."""

Base = declarative_base()


class JsonObject(TypeDecorator):
    """Decorator for objects that are stored as serialized JSON strings."""

    impl = Unicode

    def process_literal_param(self, value, dialect):
        """Expects a JSON serializable object."""
        if value is not None:
            return json.dumps(value)

    process_bind_param = process_literal_param

    def process_result_value(self, value, dialect):
        """Create JSON object from string serialization."""
        if value is not None:
            return json.loads(value)


class Archive(Base):
    """ORM for maintaining archive descriptors in a relational database."""
    # -- Schema ---------------------------------------------------------------
    __tablename__ = 'archive'

    archive_id = Column(
        String(32),
        default=util.get_unique_identifier,
        primary_key=True
    )
    name = Column(String(1024), nullable=False, unique=True)
    created_at = Column(String(32), nullable=False, default=util.current_time)
    description = Column(Text)
    encoder = Column(String(1024))
    decoder = Column(String(1024))
    serializer = Column(JsonObject)

    # -- Relationships --------------------------------------------------------
    keyspec = relationship(
        'ArchiveKey',
        back_populates='archive',
        cascade='all, delete, delete-orphan'
    )

    def descriptor(self) -> ArchiveDescriptor:
        """Get descriptor for the archive database object.

        Returns
        -------
        histore.archive.manager.descriptor.ArchiveDescriptor
        """
        # Create list of promary key columns (if defined).
        pk = [k.colid for k in sorted(self.keyspec, key=lambda x: x.pos)] if self.keyspec else None
        return ArchiveDescriptor.create(
            identifier=self.archive_id,
            name=self.name,
            description=self.description,
            primary_key=pk,
            encoder=self.encoder,
            decoder=self.decoder,
            serializer=self.serializer
        )


class ArchiveKey(Base):
    """Column in the primary key definition for an archive. Each element in the
    primary key has a unique (column) name and the key position.
    """
    # -- Schema ---------------------------------------------------------------
    __tablename__ = 'archive_key'

    archive_id = Column(
        String(32),
        ForeignKey('archive.archive_id'),
        primary_key=True
    )
    colid = Column(Integer, primary_key=True)
    pos = Column(Integer, primary_key=True)

    # -- Relationships --------------------------------------------------------
    archive = relationship('Archive', back_populates='keyspec')



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2021 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Interfaces for mixin class for dataset operators that allow to directly
process rows in a dataset snapshot to create a new snapshot.
"""

from abc import ABCMeta, abstractmethod
from typing import Dict, Optional

from histore.document.base import DataRow, InputDescriptor
from histore.document.schema import DocumentSchema


class DatasetOperator(InputDescriptor, metaclass=ABCMeta):
    """Abstract class for operators on rows in a dataset snapshots. The operator
    can be used to directly process rows in a dataset archive to create a new
    dataset snapshot.
    """
    def __init__(
        self, columns: DocumentSchema, description: Optional[str] = None,
        action: Optional[Dict] = None
    ):
        """Initialize the schema for rows that this operator will produce and
        the input descriptor.

        Parameters
        ----------
        columns: list of string
            Columns in the output schema of the operator.
        description: string, default=None
            Optional user-provided description for the snapshot that is created
            by this operator.
        action: dict, default=None
            Optional metadata defining the action that created the snapshot that
            is created by this operator.
        """
        super(DatasetOperator, self).__init__(
            description=description,
            action=action
        )
        self.columns = columns

    @abstractmethod
    def handle(self, rowid: int, row: DataRow) -> DataRow:
        """Evaluate the operator on the given row.

        Returns the processed row. If the result is None this signals that the
        given row should not be part of the collected result.

        Parameters
        -----------
        rowid: int
            Unique row identifier
        row: list
            List of values in the row.

        Returns
        -------
        list
        """
        raise NotImplementedError()  # pragma: no cover



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2021 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Writer for documents that are materialized as Json files on the file system."""

from __future__ import annotations
from datetime import datetime, date, time
from typing import Any, Optional

import json
import numpy as np


from histore.key import KeyValue

import histore.util as util


class JsonWriter(object):
    """Writer for objects that are stored as Json serialized rows in a text
    file. Each row is stored in a separate line in the text file. The
    output file is a Json array. The first and the last row of the file open
    and close the array.
    """
    def __init__(
        self, filename: str, compression: Optional[str] = None,
        encoder: Optional[json.JSONEncoder] = None
    ):
        """Initialize the output file, row counter, and the serializer that is
        being used.

        Parameters
        ----------
        filename: string
            Path to the output file.
        compression: string, default=None
            String representing the compression mode for the output file.
        encoder: json.JSONEncoder, default=None
            Encoder used when writing archive rows as JSON objects to file.
        """
        # Use the default JSONEncoder if no encoder is given
        self.encoder = encoder  # if encoder is not None else DefaultEncoder
        # Open output file for writing.
        self.fout = util.outputstream(filename, compression=compression)

    def __enter__(self) -> JsonWriter:
        """Enter method for the context manager."""
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> bool:
        """Close the document iterator when the context manager exits."""
        self.close()
        return False

    def close(self):
        """Write the last row to the output file and close the output array and
        the output file.
        """
        self.fout.close()

    def write(self, row: Optional[Any] = None):
        """Write the archive row in the internal buffer to the output file.
        Replace the buffer with the given (next output row).

        Parameters
        ----------
        row: histore.archive.row.ArchiveRow, default=None
            Next row in the output stream. This row will be kept in the
            internal buffer and the previous row is being written to the
            output file.
        """
        self.fout.write('{}\n'.format(json.dumps(row, cls=self.encoder)))


# -- Helper classes -----------------------------------------------------------

class DefaultEncoder(json.JSONEncoder):
    """Json encoder that handles numpy data types from pandas data frames and
    datetime objects.
    """
    def default(self, obj):
        """Convert numpy data types to default Python types Date objects are
        converted to dictionaries using the class an label (prefixed by the
        special character '$') and the ISO format string representation as the
        value.

        Parameters
        ----------
        obj: any
            Value that is being encoded.

        Returns
        -------
        any
        """
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        elif isinstance(obj, datetime):
            return {'$datetime': obj.isoformat()}
        elif isinstance(obj, date):
            return {'$date': obj.isoformat()}
        elif isinstance(obj, time):
            return {'$time': obj.isoformat()}
        elif isinstance(obj, KeyValue):
            assert not obj.is_new()
            return obj.value
